from copy import deepcopy
import functools
import json
import logging
import operator
import re
from tempfile import gettempdir
from urllib import parse
from urllib.parse import parse_qs, urlparse
from numpy import mean, median, std
from tabulate import tabulate
from nidm.core import Constants
from nidm.experiment import Navigate, Query
from nidm.experiment.Utils import validate_uuid


def convertListtoDict(lst):
    """
    This function converts a list to a dictionary
    :param lst: list to convert
    :return: dictionary
    """
    res_dct = {lst[i]: lst[i + 1] for i in range(0, len(lst), 2)}
    return res_dct


class RestParser:
    OBJECT_FORMAT = 0
    JSON_FORMAT = 1
    CLI_FORMAT = 2

    def __init__(self, verbosity_level=0, output_format=0):
        self.verbosity_level = verbosity_level
        self.output_format = output_format
        self.restLog(f"Setting output format {self.output_format}", 4)

    def setOutputFormat(self, output_format):
        self.output_format = output_format
        self.restLog(f"Setting output format {self.output_format}", 4)

    #####################
    # Standard formatters
    #####################

    def arrayFormat(self, result, headers):
        def allUUIDs(arr):
            uuid_only = True
            for s in arr:
                if not isinstance(s, str) or not re.match(
                    "^[0-9a-f]+-[0-9a-f]+-[0-9a-f]+-[0-9a-f]+-[0-9a-f]+$", s
                ):
                    uuid_only = False
            return uuid_only

        if self.output_format == RestParser.JSON_FORMAT:
            return json.dumps(result, indent=2)
        elif self.output_format == RestParser.CLI_FORMAT:
            # most likely this is an array of strings but tabulate wants an array of arrays
            table = []
            for s in result:
                table.append([s])
            if allUUIDs(result) and headers[0] == "":
                headers[0] = "UUID"
            return tabulate(table, headers=headers)
        return result

    def dictFormat(self, result, headers=None):
        if headers is None:
            headers = [""]
        if self.output_format == self.CLI_FORMAT:
            table = []
            appendicies = []
            for key in result:
                # format a list
                if isinstance(result[key], list):
                    appendix = []
                    for line in result[key]:
                        appendix.append([json.dumps(line)])
                    appendicies.append(tabulate(appendix, [key]))

                    # also put really short lists in as comma separated values
                    if len(json.dumps(result[key])) < 40:
                        table.append([json.dumps(key), ",".join(result[key])])

                # format a string
                elif isinstance(result[key], str):
                    table.append([json.dumps(key), result[key]])

                # format a dictionary
                elif isinstance(result[key], dict):
                    # put any dict into it's own table at the end (sort of like an appendix)
                    appendix = []
                    for inner_key in result[key]:
                        appendix.append([key, inner_key, result[key][inner_key]])
                    appendicies.append(tabulate(appendix))

                # format anything else
                else:
                    col1 = json.dumps(key)
                    if isinstance(result[key], set):
                        col2 = json.dumps(list(result[key]))
                    else:
                        col2 = json.dumps(result[key])
                    table.append([col1, col2])

            return tabulate(table, headers) + "\n\n" + "\n\n".join(appendicies)
        else:
            return self.format(result)

    def objectTableFormat(self, result, headers=None):
        def flatten(obj, maxDepth=10, table=None, rowInProgress=None, depth=0):
            if table is None:
                table = []
            if rowInProgress is None:
                rowInProgress = []
            for key in obj:
                newrow = deepcopy(rowInProgress)
                if depth < maxDepth and isinstance(obj[key], dict):
                    newrow.append(key)
                    flatten(obj[key], maxDepth, table, newrow, depth + 1)
                elif isinstance(obj[key], str):
                    newrow.append(key)
                    newrow.append(obj[key])
                    table.append(newrow)
                else:
                    newrow.append(json.dumps(obj[key]))
                    table.append(newrow)

            return table

        if headers is None:
            headers = [""]

        return tabulate(flatten(result), headers=headers)

    def activityDataTableFormat(self, data):
        headers = ["uuid", "measure", "label", "value", "unit"]
        rows = []
        for inst_or_deriv in data:
            for d in inst_or_deriv.data:
                rows.append(
                    [inst_or_deriv.uuid, d.measureOf, d.label, d.value, d.hasUnit]
                )

        return tabulate(rows, headers=headers)

    #####################
    # Custom formatters
    #####################

    def projectSummaryFormat(self, result):
        if self.output_format == self.CLI_FORMAT:
            ### added by DBK to sort things
            if "subjects" in result:
                (
                    result["subjects"]["uuid"],
                    result["subjects"]["subject id"],
                ) = self.sort_list(
                    result["subjects"]["uuid"], result["subjects"]["subject id"]
                )
            else:
                result["subjects"] = []
            if "data_elements" in result:
                (
                    result["data_elements"]["uuid"],
                    result["data_elements"]["label"],
                ) = self.sort_list(
                    result["data_elements"]["uuid"], result["data_elements"]["label"]
                )
            else:
                result["data_elements"] = []

            toptable = []
            for key in result:
                if key not in ["subjects", "data_elements", "field_values"]:
                    toptable.append([key, json.dumps(result[key])])

            if "field_values" in result and len(result["field_values"]) > 0:
                fh_header = [
                    "subject",
                    "label",
                    "value",
                    "unit",
                    "isAbout",
                ]  # result['field_values'][0].keys()
                fh_rows = [
                    [x.subject, x.label, x.value, x.hasUnit, x.isAbout]
                    for x in result["field_values"]
                ]
                field_table = tabulate(fh_rows, fh_header)
                # added by DBK, if they asked for fields then just give them the fields
                return str(field_table)
            else:
                field_table = ""

            return "{}\n\n{}\n{}\n\n{}\n{}\n\n{}".format(
                tabulate(toptable),
                ### modified by DBK to account for new dictionary format of results
                # tabulate({"subjects": result["subjects"]}, headers="keys"),
                # sort list 2 by list 1 and replace unsorted version
                tabulate([], headers=["Subject Information"]),
                tabulate(result["subjects"], headers="keys"),
                # tabulate({"data_elements": result["data_elements"]}, headers="keys"),
                tabulate([], headers=["Data Elements"]),
                tabulate(
                    {
                        "uuid": result["data_elements"]["uuid"],
                        "label": result["data_elements"]["label"],
                    },
                    headers="keys",
                ),
                field_table,
            )
        else:
            # added by DBK to check if we had fields requested then we should just return those
            if "field_values" in result:
                # convert result['field_values'] to a list for json export
                return self.format(result["field_values"])
            else:
                return self.format(result)

    def formatDerivatives(self, derivative):
        self.restLog(f"formatting derivatives in format {self.output_format}", 5)
        if self.output_format == self.CLI_FORMAT:
            table = []
            for uri in derivative:
                for measurement in derivative[uri]["values"]:
                    if measurement not in [
                        "http://www.w3.org/ns/prov#wasGeneratedBy",
                        "http://www.w3.org/1999/02/22-rdf-syntax-ns#type",
                    ]:  # skip some NIDM structure artifacts
                        table.append(
                            [
                                uri,
                                measurement,
                                derivative[uri]["values"][measurement]["label"],
                                "{} {}".format(
                                    derivative[uri]["values"][measurement]["value"],
                                    derivative[uri]["values"][measurement]["units"],
                                ),
                                derivative[uri]["values"][measurement]["datumType"],
                                derivative[uri]["values"][measurement]["isAbout"],
                            ]
                        )
            return tabulate(
                table,
                headers=[
                    "Derivative_UUID",
                    "Measurement",
                    "Label",
                    "Value",
                    "Datumtype",
                    "isAbout",
                ],
            )
        else:
            return self.format(derivative)

    def dataElementsFormat(self, de_data):
        if self.output_format == self.CLI_FORMAT:
            table = []
            headers = [
                "label",
                "source_variable",
                "hasUnit",
                "description",
                "dataElement",
                "isAbout",
            ]

            # for each data element, create a row with each value from the header
            for de in de_data["data_elements"]["data_type_info"]:
                row = []
                for h in headers:
                    row.append(de[h])
                table.append(row)

            # print (de_data['data_elements']['data_type_info'][0])

            # Sort by the first column, which is label
            sorted_table = sorted(table, key=lambda x: x[0])

            text = tabulate(sorted_table, headers)
            return text
        else:
            return self.format(de_data)

    def dataElementDetailsFormat(self, de_data):
        return self.format(de_data)

    def subjectFormat(self, subject_data):
        if self.output_format == self.CLI_FORMAT:
            subjects = []
            for subject in subject_data["subject"]:
                subjects.append(subject)
            text = tabulate(subjects, headers=["Subject UUID", "Source Subject ID"])

            if "fields" in subject_data:
                field_data = []
                text += "\n\n"
                for sub in subject_data["fields"]:
                    for act in subject_data["fields"][sub]:
                        de = subject_data["fields"][sub][act]
                        field_data.append([sub, act, de.label, de.value])
                text += tabulate(
                    field_data, headers=["Subject", "Activity", "Field", "Value"]
                )

            return text
        else:
            return self.format(subject_data)

    def subjectSummaryFormat(self, result):
        if self.output_format == self.CLI_FORMAT:
            special_keys = ["instruments", "derivatives", "activity"]
            toptable = []
            for key in result:
                if key not in special_keys:
                    toptable.append([key, result[key]])

            for key in special_keys:
                if isinstance(result[key], dict):
                    toptable.append([key, ",".join(result[key].keys())])
                elif isinstance(result[key], list):
                    toptable.append([key, ",".join(result[key])])
                else:
                    toptable.append([key, json.dumps(result[key])])

            instruments = self.objectTableFormat(
                result["instruments"], ["Instrument_UUID", "Category", "Value"]
            )
            derivatives = self.formatDerivatives(result["derivatives"])

            return f"{tabulate(toptable)}\n\n{derivatives}\n\n{instruments}"
        else:
            return self.format(result)

    def subjectSummaryFormat_v2(self, result):
        if self.output_format == self.CLI_FORMAT:
            special_keys = ["instruments", "derivatives", "activity"]
            toptable = []
            for key in result:
                if key not in special_keys:
                    toptable.append([key, result[key]])

            for key in special_keys:
                if key in result:
                    if isinstance(result[key], dict):
                        toptable.append([key, ",".join(result[key].keys())])
                    if (
                        isinstance(result[key], list)
                        and len(result[key]) > 0
                        and isinstance(result[key][0], Navigate.ActivityData)
                    ):
                        toptable.append([key, ",".join([x.uuid for x in result[key]])])
                    elif isinstance(result[key], list):
                        toptable.append([key, ",".join])
                    else:
                        toptable.append([key, json.dumps(result[key])])

            instruments = self.activityDataTableFormat(result["instruments"])
            derivatives = self.activityDataTableFormat(result["derivatives"])

            return f"{tabulate(toptable)}\n\n{derivatives}\n\n{instruments}"
        else:
            return self.format(result)

    ### Added by DBK to sorty data elements lists
    #####################
    # Sort Functions
    #####################
    def sort_list(self, list1, list2):
        """
        This function will sort list 1 using list 2 values, returning sorted list 1, sorted list 2
        """

        if len(list1) == 0 or len(list2) == 0:
            return list1, list2

        list1 = list(zip(*sorted(zip(list2, list1))))[1]
        return list1, sorted(list2)

    #####################
    # Route Functions
    #####################

    def dataelements(self):
        result = Navigate.GetDataelements(self.nidm_files)
        return self.dataElementsFormat(result)

    def dataelementsSummary(self):
        path = (urlparse(self.command)).path
        match = re.match(r"^/?dataelements/([^/\?]+)", path)
        dataelement = parse.unquote(str(match.group(1)))

        result = Navigate.GetDataelementDetails(self.nidm_files, dataelement)
        return self.dataElementDetailsFormat(result)

    def projects(self):
        result = []
        field_values = []
        self.restLog("Returning all projects", 2)
        projects = Query.GetProjectsUUID(self.nidm_files)
        for uuid in projects:
            result.append(str(uuid).replace(Constants.NIIRI, ""))

        # if we got fields, drill into each subject and pull out the field data
        # subject details -> derivatives / instrument -> values -> element
        if "fields" in self.query and len(self.query["fields"]) > 0:
            subjects_set = set()
            dataelements_set = set()
            self.restLog(f"Using fields {self.query['fields']}", 2)
            # result['field_values'] = []

            for proj in projects:
                # get all the synonyms for all the fields
                field_synonyms = functools.reduce(
                    operator.iconcat,
                    [
                        Query.GetDatatypeSynonyms(self.nidm_files, proj, x)
                        for x in self.query["fields"]
                    ],
                    [],
                )

                # files = self.nidm_files
                all_subjects = Query.GetParticipantUUIDsForProject(
                    self.nidm_files, proj, self.query["filter"]
                )  # nidm_file_list= files, project_id=proj['uuid'], filter=self.query['filter']):
                for sub in all_subjects["uuid"]:
                    for activity in Navigate.getActivities(self.nidm_files, sub):
                        activity = Navigate.getActivityData(self.nidm_files, activity)
                        for data_element in activity.data:
                            if data_element.dataElement in field_synonyms:
                                field_values.append(data_element._replace(subject=sub))
                                subjects_set.add(sub)
                                dataelements_set.add(
                                    (data_element.datumType, data_element.label)
                                )

            if len(field_values) == 0:
                raise ValueError(
                    "Supplied field not found. ("
                    + ", ".join(self.query["fields"])
                    + ")"
                )
            else:
                summary_result = {}
                summary_result["subjects"] = {"uuid": [], "subject id": []}
                for sub in subjects_set:
                    summary_result["subjects"]["uuid"].append(sub)
                    summary_result["subjects"]["subject id"].append("")
                summary_result["data_elements"] = {"uuid": [], "label": []}
                for de in dataelements_set:
                    summary_result["data_elements"]["uuid"] = de[0]
                    summary_result["data_elements"]["label"] = de[1]
                summary_result["field_values"] = field_values
                return self.projectSummaryFormat(summary_result)

        return self.format(result, ["UUID"])

    def ExpandProjectMetaData(self, meta_data):
        """
        Takes in the meta_data from GetProjectsMetadata() and adds
        the following statistics about each project to the existing
        meta_data structure:
         age_max (float)
         age_min (float)
         handedness of subjects (list)
         genders of subjects (list)
         number of subjects (int)

        :param meta_data:
        :return:
        """
        for project_id in meta_data["projects"]:
            project_uuid = (
                str(project_id)[6:]
                if (str(project_id).startswith("niiri:"))
                else project_id
            )
            project = meta_data["projects"][project_id]

            ages = set()
            hands = set()
            genders = set()

            for session in Navigate.getSessions(self.nidm_files, project_uuid):
                for acq in Navigate.getAcquisitions(self.nidm_files, session):
                    act_data = Navigate.getActivityData(self.nidm_files, acq)
                    for de in act_data.data:
                        if de.isAbout in (
                            "http://uri.interlex.org/ilx_0100400",
                            "http://uri.interlex.org/base/ilx_0100400",
                        ):
                            if de.value in ("n/a", "nan"):
                                ages.add(float("nan"))
                            else:
                                ages.add(float(de.value))
                        elif de.isAbout in (
                            "http://uri.interlex.org/ilx_0101292",
                            "http://uri.interlex.org/base/ilx_0101292",
                            "http://uri.interlex.org/ilx_0738439",
                            "https://ndar.nih.gov/api/datadictionary/v2/dataelement/gender",
                        ):
                            genders.add(de.value)
                        elif (
                            de.isAbout == "http://purl.obolibrary.org/obo/PATO_0002201"
                        ):
                            hands.add(de.value)

            print(Query.GetParticipantUUIDsForProject(self.nidm_files, project_uuid))

            project["age_max"] = max(ages) if len(ages) > 0 else 0
            project["age_min"] = min(ages) if len(ages) > 0 else 0
            project[Query.matchPrefix(str(Constants.NIDM_NUMBER_OF_SUBJECTS))] = len(
                (Query.GetParticipantUUIDsForProject(self.nidm_files, project_uuid))[
                    "uuid"
                ]
            )
            project[str(Constants.NIDM_GENDER)] = list(genders)
            project[str(Constants.NIDM_HANDEDNESS)] = list(hands)

    def projectStats(self):
        result = {}
        subjects = None
        path = (urlparse(self.command)).path

        match = re.match(r"^/?statistics/projects/([^/]+)\??$", path)
        id_ = parse.unquote(str(match.group(1)))
        self.restLog(f"Returning project {id_} stats metadata", 2)

        meta_data = Query.GetProjectsMetadata(self.nidm_files)
        self.ExpandProjectMetaData(meta_data)
        projects = Query.compressForJSONResponse(meta_data)

        for pid in projects["projects"].keys():
            self.restLog("comparng " + str(pid) + " with " + str(id_), 5)
            self.restLog("comparng " + str(pid) + " with " + Constants.NIIRI + id_, 5)
            self.restLog("comparng " + str(pid) + " with niiri:" + id_, 5)
            if pid in (id_, Constants.NIIRI + id_, "niiri:" + id_):
                # strip off prefixes to make it more human readable
                for key in projects["projects"][pid]:
                    short_key = key
                    possible_prefix = re.sub(":.*", "", short_key)
                    if possible_prefix in Constants.namespaces:
                        short_key = re.sub("^.*:", "", short_key)
                    result[short_key] = projects["projects"][pid][key]

        # now get any fields they requested
        for field in self.query["fields"]:
            if subjects is None:
                subjects = Query.GetParticipantUUIDsForProject(
                    tuple(self.nidm_files), project_id=id_, filter=self.query["filter"]
                )
                result["subjects"] = subjects["uuid"]
            bits = field.split(".")
            if len(bits) > 1:
                stat_type = self.getStatType(
                    bits[0]
                )  # should be either instruments or derivatives for now.
                self.addFieldStats(
                    result, id_, subjects["uuid"], bits[1], stat_type
                )  # bits[1] will be the ID

        return self.dictFormat(result)

    STAT_TYPE_OTHER = 0
    STAT_TYPE_INSTRUMENTS = 1
    STAT_TYPE_DERIVATIVES = 2

    def getStatType(self, name):
        lookup = {
            "instruments": self.STAT_TYPE_INSTRUMENTS,
            "derivatives": self.STAT_TYPE_DERIVATIVES,
        }
        if name in lookup:
            return lookup[name]
        return self.STAT_TYPE_OTHER

    @staticmethod
    def getTailOfURI(uri):
        if "#" in uri:
            return uri[uri.rfind("#") + 1 :]
        else:
            return uri[uri.rfind("/") + 1 :]

    def addFieldStats(self, result, project, subjects, field, type):  # noqa: A002
        """
        Geneerates basic stats on a group of subjects and adds it to the result
        :param result:
        :param subjects:
        :param field:
        :return:
        """
        values = []
        for s in subjects:
            if type == self.STAT_TYPE_INSTRUMENTS:
                data = Query.GetParticipantInstrumentData(
                    tuple(self.nidm_files), project, s
                )
                for v in data.values():
                    if field in v:
                        values.append(float(v[field]))
            # derivatives are of the form [UUID]['values'][URI]{datumType, label, values, units}
            if type == self.STAT_TYPE_DERIVATIVES:
                data = Query.GetDerivativesDataForSubject(
                    tuple(self.nidm_files), project, s
                )
                for deriv_value in data.values():
                    for URI in deriv_value["values"]:
                        measures = deriv_value["values"][URI]
                        if field == measures["label"] or field == self.getTailOfURI(
                            URI
                        ):
                            values.append(float(measures["value"]))

        if len(values) > 0:
            med = median(values)
            avg = mean(values)
            st = std(values)
            mn = min(values)
            mx = max(values)
        else:
            med = avg = st = mn = mx = None
        result[field] = {
            "max": mx,
            "min": mn,
            "median": med,
            "mean": avg,
            "standard_deviation": st,
        }

    def projectSummary(self):
        match = re.match(r"^/?projects/([^/]+)$", self.command)
        pid = parse.unquote(str(match.group(1)))
        self.restLog(f"Returning project {pid} summary", 2)

        result = Navigate.GetProjectAttributes(self.nidm_files, project_id=pid)
        result["subjects"] = Query.GetParticipantUUIDsForProject(
            self.nidm_files, project_id=pid, filter=self.query["filter"]
        )
        result["data_elements"] = Query.GetProjectDataElements(
            self.nidm_files, project_id=pid
        )

        # if we got fields, drill into each subject and pull out the field data
        # subject details -> derivatives / instrument -> values -> element
        if "fields" in self.query and len(self.query["fields"]) > 0:
            self.restLog(f"Using fields {self.query['fields']}", 2)
            result["field_values"] = []
            # get all the synonyms for all the fields
            field_synonyms = functools.reduce(
                operator.iconcat,
                [
                    Query.GetDatatypeSynonyms(self.nidm_files, pid, x)
                    for x in self.query["fields"]
                ],
                [],
            )
            for sub in result["subjects"]["uuid"]:
                for activity in Navigate.getActivities(self.nidm_files, sub):
                    activity = Navigate.getActivityData(self.nidm_files, activity)
                    for data_element in activity.data:
                        if data_element.dataElement in field_synonyms:
                            result["field_values"].append(
                                data_element._replace(subject=sub)
                            )

            if len(result["field_values"]) == 0:
                raise ValueError(
                    "Supplied field not found. ("
                    + ", ".join(self.query["fields"])
                    + ")"
                )

        return self.projectSummaryFormat(result)

    def subjectsList(self):
        match = re.match(r"^/?projects/([^/]+)/subjects/?$", self.command)
        project = match.group((1))
        self.restLog(
            f"Returning all agents matching filter '{self.query['filter']}' for project {project}",
            2,
        )
        # result = Query.GetParticipantUUIDsForProject(self.nidm_files, project, self.query['filter'], None)
        all_subjects = Navigate.getSubjects(self.nidm_files, project)
        result = {}
        result["uuid"] = []
        result["subject id"] = []
        for sub_uuid in all_subjects:
            if Query.CheckSubjectMatchesFilter(
                self.nidm_files, project, sub_uuid, self.query["filter"]
            ):
                uuid_string = (str(sub_uuid)).split("/")[
                    -1
                ]  # srip off the http://whatever/whatever/
                result["uuid"].append(uuid_string)
                sid = Navigate.getSubjectIDfromUUID(self.nidm_files, sub_uuid)
                result["subject id"].append(str(sid))
        return self.format(result)

    def projectSubjectSummary(self):
        match = re.match(r"^/?projects/([^/]+)/subjects/([^/]+)/?$", self.command)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        self.restLog(f"Returning info about subject {match[2]}", 2)
        return self.subjectSummaryFormat(
            Query.GetParticipantDetails(self.nidm_files, match.group(1), subject)
        )

    def getFieldInfoForSubject(self, project, subject):
        """
        Returns a dictionary of activities where the subject has matching field data
        The result[activity] is the full data_element so to get the value you would use result[activity].value
        Note that a subject could match the same field in multiple activities.

        :param project:
        :param subject:
        :return:
        """
        result = {}
        # if we got fields, drill into each subject and pull out the field data
        # subject details -> derivatives / instrument -> values -> element
        if "fields" in self.query and len(self.query["fields"]) > 0:
            # get all the synonyms for all the fields - we can search for them all at once
            field_synonyms = functools.reduce(
                operator.iconcat,
                [
                    Query.GetDatatypeSynonyms(self.nidm_files, project, x)
                    for x in self.query["fields"]
                ],
                [],
            )

            # print (field_synonyms)

            for activity in Navigate.getActivities(self.nidm_files, subject):
                activity_data = Navigate.getActivityData(self.nidm_files, activity)
                # print ([ x.label for x in activity.data])
                for data_element in activity_data.data:
                    if not set(
                        [
                            data_element.dataElement,
                            data_element.label,
                            data_element.isAbout,
                        ]
                    ).isdisjoint(set(field_synonyms)):
                        result[Query.URITail(activity)] = data_element
        return result

    def subjects(self):
        self.restLog("Returning info about subjects", 2)
        projects = Navigate.getProjects(self.nidm_files)
        result = {"subject": []}
        if "fields" in self.query and len(self.query["fields"]) > 0:
            result["fields"] = {}

        for proj in projects:
            subs = Navigate.getSubjects(self.nidm_files, proj)
            for s in subs:
                result["subject"].append(
                    [
                        Query.URITail(s),
                        Navigate.getSubjectIDfromUUID(self.nidm_files, s),
                    ]
                )

                # print ("getting info for " + str(s))
                x = self.getFieldInfoForSubject(proj, s)
                if x:
                    result["fields"][Query.URITail(s)] = x
        return self.subjectFormat(result)

    def subjectSummary(self):
        match = re.match(r"^/?subjects/([^/]+)/?$", self.command)
        self.restLog(f"Returning info about subject {match[1]}", 2)
        sid = match.group(1)

        # if we were passed in a sub_id rather than a UUID, lookup the associated UUID. (we might get multiple!)
        if validate_uuid(sid):
            sub_ids = sid
        else:
            sub_ids = Navigate.getSubjectUUIDsfromID(self.nidm_files, sid)
            if len(sub_ids) == 1:
                sub_ids = sub_ids[0]

        activities = Navigate.getActivities(self.nidm_files, sid)
        activityData = []
        for a in activities:
            data = Navigate.getActivityData(self.nidm_files, a)
            activityData.append(data)

        return self.subjectSummaryFormat_v2(
            {
                "uuid": sub_ids,
                "instruments": list(
                    filter(lambda x: x.category == "instrument", activityData)
                ),
                "derivatives": list(
                    filter(lambda x: x.category == "derivative", activityData)
                ),
            }
        )

    def instrumentsList(self):
        result = []
        match = re.match(
            r"^/?projects/([^/]+)/subjects/([^/]+)/instruments/?$", self.command
        )
        self.restLog(f"Returning instruments in subject {match[2]}", 2)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        instruments = Query.GetParticipantInstrumentData(
            self.nidm_files, match.group(1), subject
        )
        for i in instruments:
            result.append(i)
        return self.format(result)

    def instrumentSummary(self):
        match = re.match(
            r"^/?projects/([^/]+)/subjects/([^/]+)/instruments/([^/]+)$", self.command
        )
        self.restLog(
            f"Returning instrument {match[3]} in subject {match[2]}",
            2,
        )
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        instruments = Query.GetParticipantInstrumentData(
            self.nidm_files, match.group(1), subject
        )
        return self.format(instruments[match.group(3)], headers=["Category", "Value"])

    def derivativesList(self):
        result = []
        match = re.match(r"^/?projects/([^/]+)/subjects/([^/]+)", self.command)
        self.restLog(f"Returning derivatives in subject {match[2]}", 2)
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        derivatives = Query.GetDerivativesDataForSubject(
            self.nidm_files, match.group(1), subject
        )
        for s in derivatives:
            result.append(s)
        return self.format(result)

    def derivativeSummary(self):
        match = re.match(
            r"^/?projects/([^/]+)/subjects/([^/]+)/derivatives/([^/]+)", self.command
        )
        subject = Navigate.normalizeSingleSubjectToUUID(self.nidm_files, match.group(2))
        uri = match.group(3)
        self.restLog(f"Returning stat {uri} in subject {match[2]}", 2)
        derivatives = Query.GetDerivativesDataForSubject(
            self.nidm_files, match.group(1), subject
        )

        single_derivative = {uri: derivatives[uri]}

        self.restLog("Formatting single derivative", 5)

        return self.formatDerivatives(single_derivative)

    def run(self, nidm_files, command):
        try:
            self.restLog("parsing command " + command, 1)
            self.restLog("Files to read:" + str(nidm_files), 1)
            self.restLog(f"Using {gettempdir()} as the graph cache directory", 1)

            self.nidm_files = tuple(nidm_files)
            # replace # marks with %23 - they are sometimes used in the is_about terms
            escaped = command.replace("#", "%23")
            u = urlparse(escaped)
            self.command = u.path
            self.query = parse_qs(u.query)

            if "filter" in self.query:
                self.query["filter"] = self.query["filter"][0]
            else:
                self.query["filter"] = None

            # normalize query dict for our particular situation
            if "fields" in self.query:
                self.query["fields"] = str.split(self.query["fields"][0], ",")
            else:
                self.query["fields"] = []

            return self.route()
        except ValueError as ve:
            logging.error("Exception: %s", ve)
            return self.format(
                {"error": "One of the supplied field terms was not found."}
            )

    def route(self):
        if re.match(r"^/?dataelements/?$", self.command):
            return self.dataelements()

        if re.match(r"^/?dataelements/[^/]+/?$", self.command):
            return self.dataelementsSummary()

        if re.match(r"^/?projects/?$", self.command):
            return self.projects()

        if re.match(r"^/?statistics/projects/[^/]+$", self.command):
            return self.projectStats()

        if re.match(r"^/?projects/[^/]+$", self.command):
            return self.projectSummary()

        if re.match(r"^/?subjects/?$", self.command):
            return self.subjects()

        if re.match(r"^/?subjects/[^/]+$", self.command):
            return self.subjectSummary()

        if re.match(r"^/?projects/[^/]+/subjects/?$", self.command):
            return self.subjectsList()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/?$", self.command):
            return self.projectSubjectSummary()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/instruments/?$", self.command):
            return self.instrumentsList()

        if re.match(
            r"^/?projects/[^/]+/subjects/[^/]+/instruments/[^/]+/?$", self.command
        ):
            return self.instrumentSummary()

        if re.match(r"^/?projects/[^/]+/subjects/[^/]+/derivatives/?$", self.command):
            return self.derivativesList()

        if re.match(
            r"^/?projects/[^/]+/subjects/[^/]+/derivatives/[^/]+/?$", self.command
        ):
            return self.derivativeSummary()

        self.restLog("NO MATCH!", 2)

        return {"error": "No match for supplied URI"}

    def restLog(self, message, verbosity_of_message):
        if verbosity_of_message <= self.verbosity_level:
            print(message)

    def format(self, result, headers=None):
        if headers is None:
            headers = [""]
        if self.output_format == RestParser.JSON_FORMAT:
            json_str = json.dumps(result, indent=2)
            return json_str

        elif self.output_format == RestParser.CLI_FORMAT:
            if isinstance(result, dict):
                return self.dictFormat(result, headers)
            if isinstance(result, list):
                return self.arrayFormat(result, headers)
            else:
                return str(result)

        return result



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
import pytest
from nidm.experiment.tools.rest import RestParser
from nidm.util import urlretrieve

# We will test example NIDM files downloaded from
# the GitHub dbkeator/simple2_NIDM_examples repo
#
# DBK: this is a bit unsafe as the TTL files in the github repo above can change and the UUID will change since they are randomly
# generated at this point.  It's probably more robust to explicitly create these files for the time being and explicitly set the
# UUID in the test file:
# For example:  kwargs={Constants.NIDM_PROJECT_NAME:"FBIRN_PhaseIII",Constants.NIDM_PROJECT_IDENTIFIER:1200,Constants.NIDM_PROJECT_DESCRIPTION:"Test investigation2"}
#               project = Project(uuid="_654321",attributes=kwargs)


@pytest.fixture(scope="session")
def brain_vol_files(tmp_path_factory: pytest.TempPathFactory) -> list[str]:
    tmp_path = tmp_path_factory.mktemp("brain_vol_files")
    urlretrieve(
        # commented out until latest simple2 NIDM files contain brain volumes again...
        # "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
        "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/a7e3fa291975c08932a412d937d12199da62dd28/datasets.datalad.org/abide/RawDataBIDS/CMU_a/nidm.ttl",
        tmp_path / "cmu_a.nidm.ttl",
    )
    urlretrieve(
        # commented out until latest simple2 NIDM files contain brain volumes again...
        # "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/abide/RawDataBIDS/Caltech/nidm.ttl",
        "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/a7e3fa291975c08932a412d937d12199da62dd28/datasets.datalad.org/abide/RawDataBIDS/Caltech/nidm.ttl",
        tmp_path / "caltech.nidm.ttl",
    )
    return [
        str(tmp_path / "cmu_a.nidm.ttl"),
        str(tmp_path / "caltech.nidm.ttl"),
    ]


@dataclass
class BrainVol:
    files: list[str]
    restParser: RestParser
    cmu_test_project_uuid: str
    cmu_test_subject_uuid: str


@pytest.fixture(scope="session")
def brain_vol(brain_vol_files: list[str]) -> BrainVol:
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    projects = restParser.run(brain_vol_files, "/projects")
    cmu_test_project_uuid: Optional[str] = None
    for p in projects:
        proj_info = restParser.run(brain_vol_files, f"/projects/{p}")
        if (
            isinstance(proj_info, dict)
            and proj_info.get("dctypes:title") == "ABIDE - CMU_a"
        ):
            cmu_test_project_uuid = p
            break
    assert cmu_test_project_uuid is not None
    subjects = restParser.run(
        brain_vol_files, f"/projects/{cmu_test_project_uuid}/subjects"
    )
    cmu_test_subject_uuid = subjects["uuid"][0]
    return BrainVol(
        brain_vol_files,
        restParser,
        cmu_test_project_uuid,
        cmu_test_subject_uuid,
    )



# from nidm.core.provone import ProvONEDocument
import pytest
from nidm.core.dot import provone_to_dot

pytestmark = pytest.mark.skip(
    reason="had to comment provone import - was breaking tests from experiment"
)


@pytest.fixture(scope="module")
def doc():
    # Create new provone document with namespaces
    from nidm.core.provone import ProvONEDocument

    d1 = ProvONEDocument()
    d1.add_namespace("dcterms", "http://purl.org/dc/terms/")
    d1.add_namespace("wfms", "http://www.wfms.org/registry/")
    d1.add_namespace("rdfs", "http://www.w3.org/2000/01/rdf-schema#")
    d1.add_namespace("nowpeople", "http://www.provbook.org/nownews/people/")
    d1.add_namespace("xsd", "http://www.w3.org/2001/XMLSchema#")
    d1.add_namespace("owl", "http://www.w3.org/2002/07/owl#")

    return d1


def test_ispartof(doc, tmp_path) -> None:
    workflow_1ex1 = doc.processExec(
        "dcterms:identifier:wf1_ex1",
        "2013-08-21 13:37:54",
        "2013-08-21 13:37:59",
        {"wfms:completed": "1"},
    )
    pe1 = doc.processExec(
        "dcterms:identifier:e1_ex1", "2013-08-21 13:37:53", "2013-08-21 13:37:53"
    )

    doc.isPartOf(pe1, workflow_1ex1)

    # save a turtle file
    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(doc.serialize(format="rdf", rdf_format="ttl"))


def test_used(doc, tmp_path) -> None:
    pe1 = doc.processExec(
        "dcterms:identifier:e1_ex1", "2013-08-21 13:37:53", "2013-08-21 13:37:53"
    )
    dt1 = doc.data(
        "dcterms:identifier:defparam1",
        {
            "rdfs:label": "filename",
            "prov:value": "DLEM_NEE_onedeg_v1.0nc",
            "wfms:type": "edu.sci.wfms.basic:File",
        },
    )
    doc.used(pe1, dt1)

    # save a turtle file
    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(doc.serialize(format="rdf", rdf_format="ttl"))


def test_wasderivedfrom(doc, tmp_path) -> None:
    dt1 = doc.data(
        "dcterms:identifier:defparam1",
        {
            "rdfs:label": "filename",
            "prov:value": "DLEM_NEE_onedeg_v1.0nc",
            "wfms:type": "edu.sci.wfms.basic:File",
        },
    )
    dt2 = doc.data("dcterms:identifier:defparam2", {"rdfs:label": "filename"})
    doc.wasDerivedFrom(dt1, dt2)

    # save a turtle file
    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(doc.serialize(format="rdf", rdf_format="ttl"))


def test_dataonlink(doc, tmp_path) -> None:
    dt2 = doc.data("dcterms:identifier:defparam2", {"rdfs:label": "filename"})
    dl1 = doc.dataLink("dcterms:identifier:e1_e2DL")
    # save a turtle file
    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(doc.serialize(format="rdf", rdf_format="ttl"))
    doc.dataOnLink(dt2, dl1)


def test_wasgeneratedby(doc, tmp_path) -> None:
    dt2 = doc.data("dcterms:identifier:defparam2", {"rdfs:label": "filename"})
    pe1 = doc.processExec(
        "dcterms:identifier:e1_ex1", "2013-08-21 13:37:53", "2013-08-21 13:37:53"
    )
    doc.wasGeneratedBy(dt2, pe1)
    # save a turtle file
    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(doc.serialize(format="rdf", rdf_format="ttl"))


def test_wasassociatedwith(doc):
    pe1 = doc.processExec(
        "dcterms:identifier:e1_ex1", "2013-08-21 13:37:53", "2013-08-21 13:37:53"
    )
    p2 = doc.process("dcterms:identifier:e2", {"dcterms:title": "TemporalStatistics"})
    doc.wasAssociatedWith(pe1, p2)


def test_wasattributedto(doc):
    p2 = doc.process("dcterms:identifier:e2", {"dcterms:title": "TemporalStatistics"})
    u1 = doc.user("dcterms:identifier:John")
    doc.wasAttributedTo(p2, u1)


def test_hasinport(doc):
    p2 = doc.process("dcterms:identifier:e2", {"dcterms:title": "TemporalStatistics"})
    i1 = doc.input_port(
        "dcterms:identifier:e1_ip1",
        {
            "dcterms:title": "input_vars",
            "wfms:signature": "gov.llnl.uvcdat.cdms:CDMSVariable",
        },
    )
    doc.hasInPort(p2, i1)


def test_dltoinport(doc):
    dl1 = doc.dataLink("dcterms:identifier:e1_e2DL")
    i1 = doc.input_port(
        "dcterms:identifier:e1_ip1",
        {
            "dcterms:title": "input_vars",
            "wfms:signature": "gov.llnl.uvcdat.cdms:CDMSVariable",
        },
    )
    doc.DLToInPort(dl1, i1)


def test_documentserialize(doc, tmp_path) -> None:
    # save a turtle file
    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(doc.serialize(format="rdf", rdf_format="ttl"))


def test_write_to_dot(doc):
    dot = provone_to_dot(doc)
    dot.write_png("provone-test.png")



from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import re
import sys
from typing import Optional
import uuid
import pytest
import rdflib
from rdflib import Graph, URIRef, util
from nidm.core import Constants
from nidm.experiment import Acquisition, AssessmentObject, Project, Query, Session
from nidm.experiment.CDE import getCDEs
from nidm.experiment.tools.rest import RestParser
from nidm.util import urlretrieve
from ..conftest import BrainVol

if sys.version_info >= (3, 9):
    from importlib.resources import as_file, files
else:
    from importlib_resources import as_file, files


@dataclass
class RestTest:
    path: str
    person_uuid: str
    p2_subject_uuids: list[str]


@pytest.fixture
def rest_test(tmp_path: Path) -> RestTest:
    return makeTestFile(
        dirpath=tmp_path,
        filename="agent.ttl",
        params={"PROJECT_UUID": "p1", "PROJECT2_UUID": "p2"},
    )


@dataclass
class OpenNeuro:
    files: list[str]
    project_uri: str


@pytest.fixture(scope="module")
def openneuro(tmp_path_factory: pytest.TempPathFactory) -> OpenNeuro:
    tmp_path = tmp_path_factory.mktemp("openneuro")
    urlretrieve(
        "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/openneuro/ds000120/nidm.ttl",
        tmp_path / "ds000120.nidm.ttl",
    )
    files = [str(tmp_path / "ds000120.nidm.ttl")]
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    projects2 = restParser.run(files, "/projects")
    project_uri: Optional[str] = None
    for p in projects2:
        proj_info = restParser.run(files, f"/projects/{p}")
        if (
            "dctypes:title" in proj_info.keys()
            and proj_info["dctypes:title"]
            == "Developmental changes in brain function underlying the influence of reward processing on inhibitory control (Slot Reward)"
        ):
            project_uri = p
    assert project_uri is not None
    return OpenNeuro(files, project_uri)


def addData(acq, data):
    acq_entity = AssessmentObject(acquisition=acq)
    for key in data:
        acq_entity.add_attributes({key: data[key]})
    return acq


def makeTestFile(dirpath: Path, filename: str, params: dict) -> RestTest:
    nidm_project_name = params.get("NIDM_PROJECT_NAME", False) or "Project_name_sample"
    nidm_project_identifier = params.get("NIDM_PROJECT_IDENTIFIER", False) or 9610
    nidm_project2_identifier = params.get("NIDM_PROJECT_IDENTIFIER", False) or 550
    nidm_project_description = (
        params.get("NIDM_PROJECT_DESCRIPTION", False) or "1234356 Test investigation"
    )
    project_uuid = params.get("PROJECT_UUID", False) or "_proj1"
    project_uuid2 = params.get("PROJECT2_UUID", False) or "_proj2"
    session_uuid = params.get("SESSION_UUID", False) or "_ses1"
    session_uuid2 = params.get("SESSION2_UUID", False) or "_ses2"
    p1kwargs = {
        Constants.NIDM_PROJECT_NAME: nidm_project_name,
        Constants.NIDM_PROJECT_IDENTIFIER: nidm_project_identifier,
        Constants.NIDM_PROJECT_DESCRIPTION: nidm_project_description,
    }
    p2kwargs = {
        Constants.NIDM_PROJECT_NAME: nidm_project_name,
        Constants.NIDM_PROJECT_IDENTIFIER: nidm_project2_identifier,
        Constants.NIDM_PROJECT_DESCRIPTION: nidm_project_description,
    }

    project = Project(uuid=project_uuid, attributes=p1kwargs)
    session = Session(uuid=session_uuid, project=project)
    acq = Acquisition(uuid="_acq1", session=session)
    acq2 = Acquisition(uuid="_acq2", session=session)
    acq3 = Acquisition(uuid="_acq2", session=session)

    person = acq.add_person(attributes={Constants.NIDM_SUBJECTID: "a1_9999"})
    test_person_uuid = (str(person.identifier)).replace("niiri:", "")

    acq.add_qualified_association(person=person, role=Constants.NIDM_PARTICIPANT)

    person2 = acq2.add_person(attributes={Constants.NIDM_SUBJECTID: "a1_8888"})
    acq2.add_qualified_association(person=person2, role=Constants.NIDM_PARTICIPANT)
    person3 = acq3.add_person(attributes={Constants.NIDM_SUBJECTID: "a2_7777"})
    acq2.add_qualified_association(person=person3, role=Constants.NIDM_PARTICIPANT)

    project2 = Project(uuid=project_uuid2, attributes=p2kwargs)
    session2 = Session(uuid=session_uuid2, project=project2)
    acq4 = Acquisition(uuid="_acq3", session=session2)
    acq5 = Acquisition(uuid="_acq4", session=session2)

    person4 = acq4.add_person(attributes={Constants.NIDM_SUBJECTID: "a3_6666"})
    acq4.add_qualified_association(person=person4, role=Constants.NIDM_PARTICIPANT)
    person5 = acq5.add_person(attributes={Constants.NIDM_SUBJECTID: "a4_5555"})
    acq5.add_qualified_association(person=person5, role=Constants.NIDM_PARTICIPANT)

    # now add some assessment instrument data
    addData(
        acq,
        {
            Constants.NIDM_AGE: 9,
            Constants.NIDM_HANDEDNESS: "R",
            Constants.NIDM_DIAGNOSIS: "Anxiety",
        },
    )
    addData(
        acq2,
        {
            Constants.NIDM_AGE: 8,
            Constants.NIDM_HANDEDNESS: "L",
            Constants.NIDM_DIAGNOSIS: "ADHD",
        },
    )
    addData(
        acq4,
        {
            Constants.NIDM_AGE: 7,
            Constants.NIDM_HANDEDNESS: "A",
            Constants.NIDM_DIAGNOSIS: "Depression",
        },
    )
    addData(
        acq5,
        {
            Constants.NIDM_AGE: 6,
            Constants.NIDM_HANDEDNESS: "R",
            Constants.NIDM_DIAGNOSIS: "Depression",
        },
    )

    test_p2_subject_uuids = [
        str(person4.identifier).replace("niiri:", ""),
        str(person5.identifier).replace("niiri:", ""),
    ]

    with open(dirpath / "a.ttl", "w", encoding="utf-8") as f:
        f.write(project.graph.serialize(None, format="rdf", rdf_format="ttl"))
    with open(dirpath / "b.ttl", "w", encoding="utf-8") as f:
        f.write(project2.graph.serialize(None, format="rdf", rdf_format="ttl"))

    # create empty graph
    graph = Graph()
    for nidm_file in ("a.ttl", "b.ttl"):
        tmp = Graph()
        graph = graph + tmp.parse(
            str(dirpath / nidm_file), format=util.guess_format(str(dirpath / nidm_file))
        )

    graph.serialize(str(dirpath / filename), format="turtle")

    return RestTest(
        path=str(dirpath / filename),
        person_uuid=test_person_uuid,
        p2_subject_uuids=test_p2_subject_uuids,
    )


def test_uri_subject_list(brain_vol: BrainVol, openneuro: OpenNeuro) -> None:
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    result = restParser.run(brain_vol.files + openneuro.files, "/subjects")

    assert isinstance(result, dict)
    assert isinstance(result["subject"], list)
    assert len(result["subject"]) > 10


def test_uri_subject_list_with_fields(
    brain_vol: BrainVol, openneuro: OpenNeuro
) -> None:
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    result = restParser.run(
        brain_vol.files + openneuro.files,
        "/subjects?fields=ilx_0100400,MagneticFieldStrength",
    )  # ilx_0100400 "is about" age
    assert isinstance(result, dict)

    assert isinstance(result["subject"], list)
    assert len(result["subject"]) > 10

    assert isinstance(result["fields"], dict)
    all_fields = []
    for sub in result["fields"]:
        assert isinstance(result["fields"][sub], dict)
        for activity in result["fields"][sub]:
            all_fields.append(result["fields"][sub][activity].label)
            if result["fields"][sub][activity].value != "n/a":
                assert float(result["fields"][sub][activity].value) > 0
                assert float(result["fields"][sub][activity].value) < 125
    assert "age" in all_fields
    assert "MagneticFieldStrength" in all_fields


def test_uri_project_list(tmp_path: Path) -> None:
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    proj1_uuid = str(uuid.uuid1())
    proj2_uuid = str(uuid.uuid1())
    project = Project(uuid=proj1_uuid, attributes=kwargs)
    # save a turtle file
    with open(tmp_path / "uritest.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseIII",
        Constants.NIDM_PROJECT_IDENTIFIER: 1200,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation2",
    }
    project = Project(uuid=proj2_uuid, attributes=kwargs)
    # save a turtle file
    with open(tmp_path / "uritest2.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    restParser = RestParser()
    result = restParser.run(
        [str(tmp_path / "uritest.ttl"), str(tmp_path / "uritest2.ttl")], "/projects"
    )

    project_uuids = []

    for uuid_ in result:
        project_uuids.append(uuid_)

    assert isinstance(result, list)
    assert len(project_uuids) >= 2
    assert proj1_uuid in project_uuids
    assert proj2_uuid in project_uuids


def test_uri_project_id(openneuro: OpenNeuro) -> None:
    # try with the real brain volume files
    restParser = RestParser()
    # result = restParser.run(openneuro.files, '/projects')
    project = openneuro.project_uri
    result = restParser.run(openneuro.files, f"/projects/{project}")

    assert "dctypes:title" in result
    assert "subjects" in result
    assert len(result["subjects"]["uuid"]) > 2
    assert "data_elements" in result
    assert len(result["data_elements"]["uuid"]) > 1


def test_uri_projects_subjects_1(rest_test: RestTest) -> None:
    proj_uuid = "p2"
    restParser = RestParser()
    result = restParser.run([rest_test.path], f"/projects/{proj_uuid}/subjects")

    assert isinstance(result, dict)
    assert len(result["uuid"]) == 2

    assert rest_test.p2_subject_uuids[0] in result["uuid"]
    assert rest_test.p2_subject_uuids[1] in result["uuid"]


def test_uri_subjects(brain_vol: BrainVol) -> None:
    restParser = RestParser()
    restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
    result = restParser.run(
        brain_vol.files, f"/subjects/{brain_vol.cmu_test_subject_uuid}"
    )

    assert isinstance(result, dict)
    assert "uuid" in result
    assert "instruments" in result
    assert "derivatives" in result

    assert brain_vol.cmu_test_subject_uuid == result["uuid"]


def test_uri_projects_subjects_id(openneuro: OpenNeuro) -> None:
    restParser = RestParser()
    # result = restParser.run(openneuro.files, '/projects')
    project = openneuro.project_uri
    result = restParser.run(openneuro.files, f"/projects/{project}/subjects")
    subject = result["uuid"][0]

    uri = f"/projects/{project}/subjects/{subject}"
    result = restParser.run(openneuro.files, uri)

    assert isinstance(result, dict)
    assert result["uuid"] == subject
    assert len(result["instruments"]) > 2

    instruments = result["instruments"].values()
    all_keys = []
    for i in instruments:
        all_keys += i.keys()
    assert "age" in all_keys

    # current test data doesn't have derivatives!
    # assert len(result['derivatives']) > 0


def test_get_software_agents(brain_vol: BrainVol) -> None:
    nidm_file = brain_vol.files[0]
    rdf_graph = Query.OpenGraph(nidm_file)

    agents = Query.getSoftwareAgents(rdf_graph)

    assert len(agents) > 0

    isa = URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")

    count = 0
    for a in agents:
        for _ in rdf_graph.triples((a, isa, Constants.PROV["Agent"])):
            count += 1

    assert count == len(agents)


def test_brain_vols(brain_vol: BrainVol) -> None:
    restParser = RestParser()
    project = brain_vol.cmu_test_project_uuid
    subjects = restParser.run(brain_vol.files, f"/projects/{project}/subjects")
    subject = subjects["uuid"][0]

    data = Query.GetDerivativesDataForSubject(brain_vol.files, None, subject)

    assert len(data) > 0
    for value in data.values():
        assert "StatCollectionType" in value
        assert "URI" in value
        assert "values" in value


def test_GetParticipantDetails(brain_vol: BrainVol) -> None:
    # start = time.time()

    restParser = RestParser()
    project = brain_vol.cmu_test_project_uuid

    # start = time.time()
    subjects = restParser.run(brain_vol.files, f"/projects/{project}/subjects")
    subject = subjects["uuid"][0]

    Query.GetParticipantInstrumentData(brain_vol.files, project, subject)

    details = Query.GetParticipantDetails(brain_vol.files, project, subject)

    assert "uuid" in details
    assert "id" in details
    assert "activity" in details
    assert "instruments" in details
    assert "derivatives" in details

    # end = time.time()
    # runtime = end - start
    # assert (runtime <  4)


def test_CheckSubjectMatchesFilter(brain_vol: BrainVol) -> None:
    restParser = RestParser()
    project = brain_vol.cmu_test_project_uuid
    subjects = restParser.run(brain_vol.files, f"/projects/{project}/subjects")
    subject = subjects["uuid"][0]

    derivatives = Query.GetDerivativesDataForSubject(brain_vol.files, project, subject)

    for svalue in derivatives.values():
        for vkey in svalue["values"]:
            dt = vkey
            val = svalue["values"][vkey]["value"]
            if dt and val:
                break

    # find an actual stat and build a matching filter to make sure our matcher passes it
    filter_str = f"derivatives.{dt} eq {val}"
    assert Query.CheckSubjectMatchesFilter(
        brain_vol.files, project, subject, filter_str
    )

    instruments = Query.GetParticipantInstrumentData(brain_vol.files, project, subject)
    for inst in instruments.values():
        if "AGE_AT_SCAN" in inst:
            age = inst["AGE_AT_SCAN"]

            older = str(float(age) + 1)
            younger = str(float(age) - 1)

            assert Query.CheckSubjectMatchesFilter(
                brain_vol.files,
                project,
                subject,
                f"instruments.AGE_AT_SCAN eq {age}",
            )
            assert (
                Query.CheckSubjectMatchesFilter(
                    brain_vol.files,
                    project,
                    subject,
                    f"instruments.AGE_AT_SCAN lt {younger}",
                )
                is False
            )
            assert (
                Query.CheckSubjectMatchesFilter(
                    brain_vol.files,
                    project,
                    subject,
                    f"instruments.AGE_AT_SCAN gt {younger}",
                )
                is True
            )
            assert Query.CheckSubjectMatchesFilter(
                brain_vol.files,
                project,
                subject,
                f"instruments.AGE_AT_SCAN lt {older}",
            )
            assert (
                Query.CheckSubjectMatchesFilter(
                    brain_vol.files,
                    project,
                    subject,
                    f"instruments.AGE_AT_SCAN gt {older}",
                )
                is False
            )
        # TODO deal with spaces in identifiers and CheckSubjectMatchesFilter
        elif "age at scan" in inst:
            age = inst["age at scan"]

            older = str(float(age) + 1)
            younger = str(float(age) - 1)

            assert inst["age at scan"] is not None

            # assert Query.CheckSubjectMatchesFilter(brain_vol.files, project, subject, f"instruments.age at scan eq {age}")
            # assert (Query.CheckSubjectMatchesFilter(brain_vol.files, project, subject, f"instruments.age at scan lt {younger}") == False)
            # assert (Query.CheckSubjectMatchesFilter(brain_vol.files, project, subject, f"instruments.age at scan gt {younger}") == True)
            # assert Query.CheckSubjectMatchesFilter(brain_vol.files, project, subject, f"instruments.age at scan lt {older}")
            # assert (Query.CheckSubjectMatchesFilter(brain_vol.files, project, subject, f"instruments.age at scan gt {older}") == False)


def test_ExtremeFilters(brain_vol: BrainVol) -> None:
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    project = brain_vol.cmu_test_project_uuid

    details = restParser.run(
        brain_vol.files, f"/projects/{project}?filter=AGE_AT_SCAN gt 200"
    )
    assert len(details["subjects"]["uuid"]) == 0
    assert len(details["data_elements"]["uuid"]) > 0

    details = restParser.run(
        brain_vol.files,
        f"/projects/{project}?filter=instruments.AGE_AT_SCAN gt 0",
    )
    assert len(details["subjects"]["uuid"]) > 0
    assert len(details["data_elements"]["uuid"]) > 0


def test_Filter_Flexibility(brain_vol: BrainVol) -> None:
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    project = brain_vol.cmu_test_project_uuid

    synonyms = Query.GetDatatypeSynonyms(tuple(brain_vol.files), project, "ADOS_MODULE")
    real_synonyms = [x for x in synonyms if len(x) > 1]

    assert len(real_synonyms) > 1

    for syn in real_synonyms:
        if " " in syn:
            continue
        details = restParser.run(
            brain_vol.files, f"/projects/{project}?filter=instruments.{syn} gt 2"
        )
        assert len(details["subjects"]["uuid"]) > 0
        assert len(details["data_elements"]["uuid"]) > 0


def test_OpenGraph(brain_vol: BrainVol) -> None:
    g = Query.OpenGraph(brain_vol.files[0])
    assert isinstance(g, rdflib.graph.Graph)

    # if you call OpenGraph with something that is already a graph, it should send it back
    g2 = Query.OpenGraph(g)
    assert isinstance(g2, rdflib.graph.Graph)


def test_CDEs():
    def testrun():
        with as_file(files("nidm") / "core" / "cde_dir") as dirpath:
            graph = getCDEs(
                [str(dirpath / "ants_cde.ttl"), str(dirpath / "fs_cde.ttl")]
            )

        units = graph.objects(
            subject=Constants.FREESURFER["fs_000002"],
            predicate=Constants.NIDM["hasUnit"],
        )
        count = 0
        for u in units:
            count += 1
            assert str(u) == "mm^2"

        assert count == 1

    testrun()
    getCDEs.cache = None  # clear the memory cache and try again
    testrun()  # run a second time to test disk caching.


def assess_one_col_output(txt_output):
    # print (txt_output)
    lines = txt_output.strip().splitlines()
    while not re.search(
        "[a-zA-Z]", lines[0]
    ):  # sometimes we get a blank main table, that is ok, just remove it and look at the next table
        lines = lines[1:]
    if not (re.search("UUID", lines[0]) or re.search("uuid", lines[0])):
        print(lines)
    assert re.search("UUID", lines[0]) or re.search("uuid", lines[0])
    # assert re.search('^-+$', lines[1])
    # added by DBK to deal with varying line numbers for uuids depending on the rest query type
    for line in lines:
        if is_uuid(line.strip('"')):
            return line.strip('"')
    # if we didn't find a line with a uuid then we simply flag a false assertion and return the first line of output
    # cause it doesn't really matter at this point the assertion already failed
    raise AssertionError


def is_uuid(uuid):
    return (
        re.search("^[0-9a-z]+-[0-9a-z]+-[0-9a-z]+-[0-9a-z]+-[0-9a-z]+$", uuid)
        is not None
    )


def test_cli_rest_routes(brain_vol: BrainVol) -> None:
    rest_parser = RestParser(verbosity_level=0)
    rest_parser.setOutputFormat(RestParser.CLI_FORMAT)

    #
    # / projects
    #

    text = rest_parser.run(brain_vol.files, "/projects")
    project_uuid = assess_one_col_output(text)

    #
    # /statistics/projects/{}
    #

    txt_out = rest_parser.run(brain_vol.files, f"/statistics/projects/{project_uuid}")
    lines = txt_out.strip().splitlines()
    assert re.search("^-+ +-+$", lines[0])
    lines = lines[1:]  # done testing line one, slice it off

    split_lines = [x.split() for x in lines]
    found_gender = found_age_max = found_age_min = found_title = False
    for split in split_lines:
        if len(split) > 0:  # skip blank lines between apendicies
            if re.search("title", split[0]):
                found_title = True
            if re.search("age_max", split[0]):
                found_age_max = True
            if re.search("age_min", split[0]):
                found_age_min = True
            if re.search("gender", split[0]):
                found_gender = True

    assert found_title
    assert found_age_max
    assert found_age_min
    assert found_gender

    #
    # /projects/{}/subjects
    #

    sub_text = rest_parser.run(brain_vol.files, f"/projects/{project_uuid}/subjects")
    subject_uuid = assess_one_col_output(sub_text)

    #
    # /projects/{}/subjects/{}/instruments
    #
    # result should be in 3 sections: summary , derivatives, instruments

    inst_text = rest_parser.run(
        brain_vol.files,
        f"/projects/{project_uuid}/subjects/{subject_uuid}/",
    )
    sections = inst_text.split("\n\n")

    # summary tests
    summary_lines = (
        sections[0].strip().splitlines()[1:-1]
    )  # first and last lines should be -----
    summary = {}
    for ln in summary_lines:
        summary[ln.split()[0]] = ln.split()[1]
    inst_uuid = summary["instruments"].split(",")[0]
    deriv_uuid = summary["derivatives"].split(",")[0]
    assert is_uuid(inst_uuid)
    assert is_uuid(deriv_uuid)

    # derivatives test
    deriv_lines = sections[1].strip().splitlines()
    deriv_headers = deriv_lines[0].split()
    heads = ["Derivative_UUID", "Measurement", "Label", "Value", "Datumtype"]
    for i in range(len(heads)):
        assert re.search(heads[i], deriv_headers[i], re.IGNORECASE)
    d_uuid = deriv_lines[2].split()[0]
    assert is_uuid(d_uuid)
    assert d_uuid in summary["derivatives"].split(",")

    # instruments test
    inst_lines = sections[2].strip().splitlines()
    inst_headers = inst_lines[0].split()
    heads = ["Instrument_UUID", "Category", "Value"]
    for i in range(len(heads)):
        assert re.search(heads[i], inst_headers[i], re.IGNORECASE)
    i_uuid = inst_lines[2].split()[0]
    assert is_uuid(i_uuid)
    assert i_uuid in summary["instruments"].split(",")


def test_multiple_project_fields(brain_vol: BrainVol) -> None:
    rest_parser = RestParser(verbosity_level=0)
    # rest_parser.setOutputFormat(RestParser.CLI_FORMAT)
    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = "fs_000003,ilx_0100400"  # ilx0100400 is 'isAbout' age
    fields = rest_parser.run(brain_vol.files, f"/projects?fields={field}")

    # edited by DBK to account for only field values being returned
    # assert( 'field_values' in project )
    assert len(fields) > 0
    # fv = project['field_values']
    print(fields)
    fv = fields
    assert isinstance(fv, list)
    fields_used = {i.label for i in fv}
    assert ("brain" in fields_used) or (
        "Brain Segmentation Volume (mm^3)" in fields_used
    )
    assert "age at scan" in fields_used


def test_odd_isabout_uris(brain_vol: BrainVol) -> None:
    rest_parser = RestParser(verbosity_level=0)
    # rest_parser.setOutputFormat(RestParser.CLI_FORMAT)
    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = "http://www.cognitiveatlas.org/ontology/cogat.owl#CAO_00962"
    fields = rest_parser.run(brain_vol.files, f"/projects?fields={field}")

    # edited by DBK to account for only field values being returned
    # assert( 'field_values' in project )
    assert len(fields) > 0
    # fv = project['field_values']
    print(fields)
    fv = fields
    assert isinstance(fv, list)
    fields_used = {i.label for i in fv}
    assert "ADOS_TOTAL" in fields_used


def test_project_fields_deriv(brain_vol: BrainVol) -> None:
    rest_parser = RestParser(verbosity_level=0)
    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = "fs_000003"
    project = rest_parser.run(
        brain_vol.files, f"/projects/{brain_vol.cmu_test_project_uuid}?fields={field}"
    )

    # edited by DBK to account for only field values being returned
    # assert( 'field_values' in project )
    assert len(project) > 0
    # fv = project['field_values']
    fv = project
    assert isinstance(fv, list)
    fields_used = {i.label for i in fv}
    assert ("brain" in fields_used) or (
        "Brain Segmentation Volume (mm^3)" in fields_used
    )


def test_project_fields_instruments(brain_vol: BrainVol) -> None:
    rest_parser = RestParser(verbosity_level=0)

    proj_uuid = brain_vol.cmu_test_project_uuid

    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = "age at scan"
    uri = f"/projects/{proj_uuid}?fields={field}"
    project = rest_parser.run(brain_vol.files, uri)

    # edited by DBK to account for only field values being returned
    # assert( 'field_values' in project )
    assert len(project) > 0
    # fv = project['field_values']
    fv = project
    assert isinstance(fv, list)
    fields_used = {i.label for i in fv}
    assert field in fields_used


def test_project_fields_not_found(brain_vol: BrainVol) -> None:
    # test that things don't break if the field isn't in project
    rest_parser = RestParser(verbosity_level=0)
    rest_parser.setOutputFormat(RestParser.OBJECT_FORMAT)

    field = "not_real_field"
    project = rest_parser.run(
        brain_vol.files, f"/projects/{brain_vol.cmu_test_project_uuid}?fields={field}"
    )
    keys = set(project)
    assert "error" in keys


# ATC - fail
def test_GetProjectsComputedMetadata(brain_vol: BrainVol) -> None:
    rest = RestParser()
    rest.nidm_files = tuple(brain_vol.files)
    meta_data = Query.GetProjectsMetadata(brain_vol.files)
    rest.ExpandProjectMetaData(meta_data)
    parsed = Query.compressForJSONResponse(meta_data)

    p3: Optional[str] = None
    for project_id in parsed["projects"]:
        if (
            parsed["projects"][project_id][str(Constants.NIDM_PROJECT_NAME)]
            == "ABIDE - CMU_a"
        ):
            p3 = project_id
            break
    assert p3 is not None
    assert parsed["projects"][p3][str(Constants.NIDM_PROJECT_NAME)] == "ABIDE - CMU_a"
    assert (
        parsed["projects"][p3][
            Query.matchPrefix(str(Constants.NIDM_NUMBER_OF_SUBJECTS))
        ]
        == 14
    )
    # assert parsed['projects'][p3]["age_min"] == 21.0
    # assert parsed['projects'][p3]["age_max"] == 33.0
    assert set(parsed["projects"][p3][str(Constants.NIDM_GENDER)]) == set(["1", "2"])



""" Tools for working with NIDM-Experiment files """

from argparse import ArgumentParser
import os.path
from rdflib import Graph, util
from nidm.experiment.Utils import read_nidm


def main():
    parser = ArgumentParser(
        description="This program contains various NIDM-Experiment utilities"
    )
    sub = parser.add_subparsers(dest="command")
    concat = sub.add_parser(
        "concat",
        description="This command will simply concatenate the supplied NIDM files into a single output",
    )
    visualize = sub.add_parser(
        "visualize",
        description="This command will produce a visualization(pdf) of the supplied NIDM files",
    )
    jsonld = sub.add_parser(
        "jsonld", description="This command will save NIDM files as jsonld"
    )

    for arg in [concat, visualize, jsonld]:
        arg.add_argument(
            "-nl",
            "--nl",
            dest="nidm_files",
            nargs="+",
            required=True,
            help="A comma separated list of NIDM files with full path",
        )

    concat.add_argument(
        "-o",
        "--o",
        dest="output_file",
        required=True,
        help="Merged NIDM output file name + path",
    )
    # visualize.add_argument('-o', '--o', dest='output_file', required=True, help="Output file name+path of dot graph")

    args = parser.parse_args()

    # concatenate nidm files
    if args.command == "concat":
        # create empty graph
        graph = Graph()
        for nidm_file in args.nidm_files:
            tmp = Graph()
            graph = graph + tmp.parse(nidm_file, format=util.guess_format(nidm_file))

        graph.serialize(args.output_file, format="turtle")

    elif args.command == "visualize":
        for nidm_file in args.nidm_files:
            # read in nidm file
            project = read_nidm(nidm_file)

            # split path and filename for output file writing
            file_parts = os.path.split(nidm_file)

            # write graph as nidm filename + .pdf
            project.save_DotGraph(
                filename=os.path.join(
                    file_parts[0], os.path.splitext(file_parts[1])[0] + ".pdf"
                ),
                format="pdf",
            )

        # create empty graph
        # graph=Graph()
        # for nidm_file in args.nidm_files:
        #     tmp = Graph()
        #     graph = graph + tmp.parse(nidm_file,format=util.guess_format(nidm_file))

        # project=read_nidm(StringIO.write(graph.serialize(format='turtle')))
        # project.save_DotGraph(filename=args.output_file+'.pdf',format='pdf')
        # WIP: Workaround because not all NIDM files only contain NIDM-E objects and so read_nidm function needs to be
        # updated for project.save_DotGraph to work...so this is a clunky workaround using the command line tool
        # rdf2dot

        # result is the standard output dot graph stream
        # write temporary file to disk and use for stats
        # temp = tempfile.NamedTemporaryFile(delete=False)
        # temp.write(graph.serialize(format='turtle'))
        # temp.close()
        # uber_nidm_file = temp.name
        # result = subprocess.run(['rdf2dot',uber_nidm_file], stdout=subprocess.PIPE)

        # now use graphviz Source to create dot graph object
        # src=Source(result)
        # src.render(args.output_file+'.pdf',view=False,format='pdf')

    elif args.command == "jsonld":
        # create empty graph
        for nidm_file in args.nidm_files:
            project = read_nidm(nidm_file)
            # serialize to jsonld
            with open(
                os.path.splitext(nidm_file)[0] + ".json", "w", encoding="utf-8"
            ) as f:
                f.write(project.serializeJSONLD())


if __name__ == "__main__":
    main()



import prov.model as pm
from .Core import Core, getUUID
from ..core import Constants


class Acquisition(pm.ProvActivity, Core):
    """Class for NIDM-Experiment Acquisition-Level Objects.

    Default constructor uses empty graph with namespaces added from NIDM/Scripts/Constants.py.
    Additional alternate constructors for user-supplied graphs and default namespaces (i.e. from Constants.py)
    and user-supplied graph and namespaces

    @author: David Keator <dbkeator@uci.edu>
    @copyright: University of California, Irvine 2017

    """

    # constructor
    def __init__(self, session, attributes=None, uuid=None, add_default_type=True):
        """
        Default constructor, creates a session activity and links to project object

        :param session: a session object
        :param uuid: optional uuid...used mostly for reading in existing NIDM document
        :param attributes: optional dictionary of attributes to add qname:value

        """

        # carry graph object around
        self.graph = session.graph

        if uuid is None:
            self._uuid = getUUID()

            # execute default parent class constructor
            super().__init__(
                session.graph,
                pm.QualifiedName(
                    pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                ),
                attributes,
            )
        else:
            # check if niiri namespace is already defined
            niiri_ns = self.find_namespace_with_uri(uri=str(Constants.NIIRI))
            self._uuid = uuid

            if niiri_ns is False:
                super().__init__(
                    session.graph,
                    pm.QualifiedName(
                        pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                    ),
                    attributes,
                )
            else:
                super().__init__(
                    session.graph,
                    pm.QualifiedName(niiri_ns, self.get_uuid()),
                    attributes,
                )

        session.graph._add_record(self)

        if add_default_type:
            self.add_attributes({pm.PROV_TYPE: Constants.NIDM_ACQUISITION_ACTIVITY})
        # self.add_attributes({pm.QualifiedName(pm.Namespace("dct",Constants.DCT),'isPartOf'):self})

        # list to store acquisition objects associated with this activity
        self._acquisition_objects = []
        # if constructor is called with a session object then add this acquisition to the session

        # add acquisition to session
        session.add_acquisition(self)

    def add_acquisition_object(self, acquisition_object):
        """
        Adds acquisition objects to acquisition activity, creating links and adding reference to acquisitions list

        :param acquisition: object of type "AcquisitionObject" from nidm API

        """
        # add acquisition object to self._acquisitions list
        self._acquisition_objects.extend([acquisition_object])
        # create links in graph
        self.graph.wasGeneratedBy(acquisition_object, self)

    def get_acquisition_objects(self):
        return self._acquisition_objects

    def acquisition_object_exists(self, uuid):
        """
        Checks whether uuid is a registered acquisition object
        :param uuid: full uuid of acquisition
        :return: True if exists, False otherwise
        """
        return bool(uuid in self._acquisition_objects)

    def __str__(self):
        return "NIDM-Experiment Acquisition Class"



from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import tempfile
from typing import Optional
import prov.model as pm
import pytest
from nidm.core import Constants
from nidm.experiment import (
    Acquisition,
    AssessmentAcquisition,
    AssessmentObject,
    Project,
    Query,
    Session,
)
from nidm.experiment.CDE import download_cde_files
import nidm.experiment.Navigate


@dataclass
class ProjectData:
    files: list[str]
    cmu_test_project_uuid: str


@pytest.fixture(scope="module", autouse="True")
def abide(brain_vol_files) -> ProjectData:
    files = [f for f in brain_vol_files if Path(f).name == "cmu_a.nidm.ttl"]
    assert files
    projects = Query.GetProjectsUUID(files)
    cmu_test_project_uuid: Optional[str] = None
    for p in projects:
        proj_info = nidm.experiment.Navigate.GetProjectAttributes(files, p)
        if (
            "dctypes:title" in proj_info
            and proj_info["dctypes:title"] == "ABIDE - CMU_a"
        ):
            cmu_test_project_uuid = p
            break
    assert cmu_test_project_uuid is not None
    return ProjectData(files, cmu_test_project_uuid)


def test_GetProjectMetadata(tmp_path: Path) -> None:
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    project = Project(uuid="_123456", attributes=kwargs)

    # save a turtle file
    with open(tmp_path / "test_gpm.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseIII",
        Constants.NIDM_PROJECT_IDENTIFIER: 1200,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation2",
    }
    project = Project(uuid="_654321", attributes=kwargs)

    # save a turtle file
    with open(tmp_path / "test2_gpm.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    # WIP test = Query.GetProjectMetadata(["test.ttl", "test2.ttl"])

    # assert URIRef(Constants.NIDM + "_654321") in test
    # assert URIRef(Constants.NIDM + "_123456") in test
    # assert URIRef(Constants.NIDM_PROJECT_IDENTIFIER + "1200") in test
    # assert URIRef(Constants.NIDM_PROJECT_IDENTIFIER + "9610") in test
    # assert URIRef((Constants.NIDM_PROJECT_NAME + "FBIRN_PhaseII")) in test
    # assert URIRef((Constants.NIDM_PROJECT_NAME + "FBIRN_PhaseIII")) in test
    # assert URIRef((Constants.NIDM_PROJECT_DESCRIPTION + "Test investigation")) in test
    # assert URIRef((Constants.NIDM_PROJECT_DESCRIPTION + "Test investigation2")) in test


def test_GetProjects(tmp_path: Path) -> None:
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    project = Project(uuid="_123456", attributes=kwargs)

    # save a turtle file
    with open(tmp_path / "test_gp.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    project_list = Query.GetProjectsUUID([str(tmp_path / "test_gp.ttl")])

    assert Constants.NIIRI + "_123456" in [str(x) for x in project_list]


def test_GetParticipantIDs(tmp_path: Path) -> None:
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    project = Project(uuid="_123456", attributes=kwargs)
    session = Session(uuid="_13579", project=project)
    acq = Acquisition(uuid="_15793", session=session)
    acq2 = Acquisition(uuid="_15795", session=session)

    person = acq.add_person(attributes={Constants.NIDM_SUBJECTID: "9999"})
    acq.add_qualified_association(person=person, role=Constants.NIDM_PARTICIPANT)

    person2 = acq2.add_person(attributes={Constants.NIDM_SUBJECTID: "8888"})
    acq2.add_qualified_association(person=person2, role=Constants.NIDM_PARTICIPANT)

    # save a turtle file
    with open(tmp_path / "test_3.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    participant_list = Query.GetParticipantIDs([str(tmp_path / "test_3.ttl")])

    assert participant_list["ID"].str.contains("9999").any()
    assert participant_list["ID"].str.contains("8888").any()


def test_GetProjectInstruments(tmp_path: Path) -> None:
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    proj_uuid = "_123456gpi"
    project = Project(uuid=proj_uuid, attributes=kwargs)

    session = Session(project)
    acq = AssessmentAcquisition(session)

    kwargs = {
        pm.PROV_TYPE: pm.QualifiedName(
            pm.Namespace("nidm", Constants.NIDM), "NorthAmericanAdultReadingTest"
        )
    }
    AssessmentObject(acq, attributes=kwargs)

    acq2 = AssessmentAcquisition(session)

    kwargs = {
        pm.PROV_TYPE: pm.QualifiedName(
            pm.Namespace("nidm", Constants.NIDM), "PositiveAndNegativeSyndromeScale"
        )
    }
    AssessmentObject(acq2, attributes=kwargs)

    # save a turtle file
    with open(tmp_path / "test_gpi.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    assessment_list = Query.GetProjectInstruments(
        [str(tmp_path / "test_gpi.ttl")], proj_uuid
    )

    assert Constants.NIDM + "NorthAmericanAdultReadingTest" in [
        str(x) for x in assessment_list["assessment_type"].to_list()
    ]
    assert Constants.NIDM + "PositiveAndNegativeSyndromeScale" in [
        str(x) for x in assessment_list["assessment_type"].to_list()
    ]


"""
The test data file could/should have the following project meta data. Taken from
https://raw.githubusercontent.com/incf-nidash/nidm/master/nidm/nidm-experiment/terms/nidm-experiment.owl

  - description
  - fileName
  - license
  - source
  - title
  - hadNumericalValue ???
  - BathSolution ???
  - CellType
  - ChannelNumber
  - ElectrodeImpedance
  - GroupLabel
  - HollowElectrodeSolution
  - hadImageContrastType
  - hadImageUsageType
  - NumberOfChannels
  - AppliedFilter
  - SolutionFlowSpeed
  - RecordingLocation

Returns the

"""


def saveTestFile(file_name, data):
    project = Project(uuid="_123_" + file_name, attributes=data)

    return saveProject(file_name, project)


def saveProject(file_name, project):
    # save a turtle file
    with open(file_name, "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())
    return f"nidm:_123_{file_name}"


def makeProjectTestFile(filename):
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",  # this is the "title"
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
        Constants.NIDM_FILENAME: "testfile.ttl",
        Constants.NIDM_PROJECT_LICENSE: "MIT Licence",
        Constants.NIDM_PROJECT_SOURCE: "Educational Source",
        Constants.NIDM_HAD_NUMERICAL_VALUE: "numval???",
        Constants.NIDM_BATH_SOLUTION: "bath",
        Constants.NIDM_CELL_TYPE: "ctype",
        Constants.NIDM_CHANNEL_NUMBER: "5",
        Constants.NIDM_ELECTRODE_IMPEDANCE: ".01",
        Constants.NIDM_GROUP_LABEL: "group 123",
        Constants.NIDM_HOLLOW_ELECTRODE_SOLUTION: "water",
        Constants.NIDM_HAD_IMAGE_CONTRACT_TYPE: "off",
        Constants.NIDM_HAD_IMAGE_USAGE_TYPE: "abcd",
        Constants.NIDM_NUBMER_OF_CHANNELS: "11",
        Constants.NIDM_APPLIED_FILTER: "on",
        Constants.NIDM_SOLUTION_FLOW_SPEED: "2.8",
        Constants.NIDM_RECORDING_LOCATION: "lab",
    }
    return saveTestFile(filename, kwargs)


def makeProjectTestFile2(filename):
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "TEST B",  # this is the "title"
        Constants.NIDM_PROJECT_IDENTIFIER: 1234,
        Constants.NIDM_PROJECT_DESCRIPTION: "More Scans",
        Constants.NIDM_FILENAME: "testfile2.ttl",
        Constants.NIDM_PROJECT_LICENSE: "Creative Commons",
        Constants.NIDM_PROJECT_SOURCE: "Other",
        Constants.NIDM_HAD_NUMERICAL_VALUE: "numval???",
        Constants.NIDM_BATH_SOLUTION: "bath",
        Constants.NIDM_CELL_TYPE: "ctype",
        Constants.NIDM_CHANNEL_NUMBER: "5",
        Constants.NIDM_ELECTRODE_IMPEDANCE: ".01",
        Constants.NIDM_GROUP_LABEL: "group 123",
        Constants.NIDM_HOLLOW_ELECTRODE_SOLUTION: "water",
        Constants.NIDM_HAD_IMAGE_CONTRACT_TYPE: "off",
        Constants.NIDM_HAD_IMAGE_USAGE_TYPE: "abcd",
        Constants.NIDM_NUBMER_OF_CHANNELS: "11",
        Constants.NIDM_APPLIED_FILTER: "on",
        Constants.NIDM_SOLUTION_FLOW_SPEED: "2.8",
        Constants.NIDM_RECORDING_LOCATION: "lab",
    }
    project = Project(uuid="_123_" + filename, attributes=kwargs)
    s1 = Session(project)

    a1 = AssessmentAcquisition(session=s1)
    # = s1.add_acquisition("a1", attributes={"http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#Age" : 22})

    p1 = a1.add_person(
        "p1", attributes={Constants.NIDM_GIVEN_NAME: "George", Constants.NIDM_AGE: 22}
    )
    a1.add_qualified_association(person=p1, role=Constants.NIDM_PARTICIPANT)

    return saveProject(filename, project)


def test_GetProjectsMetadata(abide: ProjectData, tmp_path: Path) -> None:
    p1 = makeProjectTestFile(str(tmp_path / "testfile.ttl"))
    p2 = makeProjectTestFile2(str(tmp_path / "testfile2.ttl"))
    files = [
        str(tmp_path / "testfile.ttl"),
        str(tmp_path / "testfile2.ttl"),
        *abide.files,
    ]

    parsed = Query.GetProjectsMetadata(files)

    # assert parsed['projects'][p1][str(Constants.NIDM_PROJECT_DESCRIPTION)] == "Test investigation"
    # assert parsed['projects'][p2][str(Constants.NIDM_PROJECT_DESCRIPTION)] == "More Scans"

    # we shouldn't have the computed metadata in this result
    # assert parsed['projects'][p1].get (Query.matchPrefix(str(Constants.NIDM_NUMBER_OF_SUBJECTS)), -1) == -1

    # find the project ID from the CMU file
    p3 = None
    for project_id in parsed["projects"]:
        if project_id not in (p1, p2):
            if (
                parsed["projects"][project_id][str(Constants.NIDM_PROJECT_NAME)]
                == "ABIDE - CMU_a"
            ):
                p3 = project_id
                break
    assert p3 is not None


def test_prefix_helpers():
    assert (
        Query.expandNIDMAbbreviation("ndar:src_subject_id")
        == "https://ndar.nih.gov/api/datadictionary/v2/dataelement/src_subject_id"
    )

    assert Query.matchPrefix("http://purl.org/nidash/nidm#abc") == "nidm:abc"
    assert Query.matchPrefix("http://www.w3.org/ns/prov#123") == "prov:123"
    assert Query.matchPrefix("http://purl.org/nidash/fsl#xyz") == "fsl:xyz"
    assert Query.matchPrefix("http://purl.org/nidash/fsl#xyz", short=True) == "fsl"


def test_getProjectAcquisitionObjects(abide: ProjectData) -> None:
    files = abide.files

    project_list = Query.GetProjectsUUID(files)
    project_uuid = str(project_list[0])
    objects = Query.getProjectAcquisitionObjects(files, project_uuid)

    assert isinstance(objects, list)


def test_GetProjectAttributes(abide: ProjectData) -> None:
    files = abide.files

    project_uuid = abide.cmu_test_project_uuid
    project_attributes = nidm.experiment.Navigate.GetProjectAttributes(
        files, project_uuid
    )
    # 8/31/23 - commented out test because we remove local prov:Location if datalad remotes aren't available
    # so this test will fail under those circumstances
    # assert ("prov:Location" in project_attributes) or ("Location" in project_attributes)
    assert ("dctypes:title" in project_attributes) or ("title" in project_attributes)
    assert (
        "http://www.w3.org/1999/02/22-rdf-syntax-ns#type" in project_attributes
    ) or ("type" in project_attributes)
    assert "AcquisitionModality" in project_attributes
    assert "ImageContrastType" in project_attributes
    assert "Task" in project_attributes
    assert "ImageUsageType" in project_attributes


def test_download_cde_files():
    cde_dir = download_cde_files()
    assert cde_dir == tempfile.gettempdir()
    fcount = 0
    for url in Constants.CDE_FILE_LOCATIONS:
        fname = url.split("/")[-1]
        assert Path(cde_dir, fname).is_file()
        fcount += 1
    assert fcount > 0


@pytest.mark.skip(
    reason="We don't have an easily accessible file for this test so skipping it until better test samples are available."
)
def test_custom_data_types():
    SPECIAL_TEST_FILES = ["/opt/project/ttl/MTdemog_aseg.ttl"]

    valuetype1 = Query.getDataTypeInfo(
        Query.OpenGraph(SPECIAL_TEST_FILES[0]), "no-real-value"
    )
    assert valuetype1 is False

    valuetype2 = Query.getDataTypeInfo(
        Query.OpenGraph(SPECIAL_TEST_FILES[0]), Constants.NIIRI["age_e3hrcc"]
    )
    assert str(valuetype2["label"]) == "age"
    assert str(valuetype2["description"]) == "Age of participant at scan"
    assert str(valuetype2["isAbout"]) == str(Constants.NIIRI["24d78sq"])

    valuetype3 = Query.getDataTypeInfo(
        Query.OpenGraph(SPECIAL_TEST_FILES[0]), "age_e3hrcc"
    )
    assert str(valuetype3["label"]) == "age"
    assert str(valuetype3["description"]) == "Age of participant at scan"
    assert str(valuetype3["isAbout"]) == str(Constants.NIIRI["24d78sq"])



import prov.model as pm
from .Core import Core, getUUID
from ..core import Constants


class Session(pm.ProvActivity, Core):
    """Class for NIDM-Experimenent Session-Level Objects.

    Default constructor uses empty graph with namespaces added from NIDM/Scripts/Constants.py.
    Additional alternate constructors for user-supplied graphs and default namespaces (i.e. from Constants.py)
    and user-supplied graph and namespaces

    @author: David Keator <dbkeator@uci.edu>
    @copyright: University of California, Irvine 2017

    """

    # constructor
    def __init__(self, project, uuid=None, attributes=None, add_default_type=True):
        """
        Default constructor, creates a session activity and links to project object

        :param project: a project object
        :return: none

        """
        if uuid is None:
            self._uuid = getUUID()
            # execute default parent class constructor
            super().__init__(
                project.graph,
                pm.QualifiedName(
                    pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                ),
                attributes,
            )
        else:
            self._uuid = uuid
            # execute default parent class constructor
            super().__init__(
                project.graph,
                pm.QualifiedName(
                    pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                ),
                attributes,
            )

        project.graph._add_record(self)

        if add_default_type:
            self.add_attributes({pm.PROV_TYPE: Constants.NIDM_SESSION})

        self.graph = project.graph
        project.add_sessions(self)

        # list of acquisitions associated with this session
        self._acquisitions = []

    def add_acquisition(self, acquisition):
        self._acquisitions.extend([acquisition])
        # create links in graph
        acquisition.add_attributes(
            {pm.QualifiedName(pm.Namespace("dct", Constants.DCT), "isPartOf"): self}
        )

    def get_acquisitions(self):
        return self._acquisitions

    def acquisition_exist(self, uuid):
        """
        Checks whether uuid is a registered acquisition
        :param uuid: full uuid of acquisition
        :return: True if exists, False otherwise
        """
        # print(f"Query uuid: {uuid}")
        for acquisitions in self._acquisitions:
            # print(acquisitions._identifier._localpart)
            if str(uuid) == acquisitions._identifier._localpart:
                return True

        return False

    def __str__(self):
        return "NIDM-Experiment Session Class"



from nidm.core import Constants
from nidm.experiment import (
    AssessmentAcquisition,
    AssessmentObject,
    DemographicsObject,
    MRAcquisition,
    MRObject,
    Project,
    Session,
)


# dj TODO: adding more tests; I only put the Dave's pipeline to a function
def main():
    # create new nidm-experiment document with project
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "Test Project name",
        Constants.NIDM_PROJECT_IDENTIFIER: "123456",
        Constants.NIDM_PROJECT_DESCRIPTION: "Test Project Description",
    }
    project = Project(attributes=kwargs)

    # test add string attribute with existing namespace
    # nidm_doc.addLiteralAttribute("nidm","isFun","ForMe")
    # project.add_attributes({Constants.PROV["Location"]:"http://nidm.nidash.org/"})

    # test add PI to investigation
    project_PI = project.add_person(
        attributes={
            Constants.NIDM_FAMILY_NAME: "Doe",
            Constants.NIDM_GIVEN_NAME: "John",
        }
    )

    # add qualified association of project PI to project activity
    project.add_qualified_association(person=project_PI, role=Constants.NIDM_PI)

    # test add session to graph and associate with project
    session = Session(project)
    session.add_attributes({Constants.NIDM_DESCRIPTION: "test session activity"})

    # test add MR acquisition activity / entity to graph and associate with session
    acq_act = MRAcquisition(session=session)
    # test add acquisition object entity to graph associated with participant role NIDM_PARTICIPANT
    acq_entity = MRObject(acquisition=acq_act)

    # add person to graph
    person = acq_act.add_person(attributes={Constants.NIDM_GIVEN_NAME: "George"})
    # add qualified association of person with role NIDM_PARTICIPANT, and associated with acquisition activity
    acq_act.add_qualified_association(person=person, role=Constants.NIDM_PARTICIPANT)

    # test add Assessment acquisition activity / entity to graph and associate with session
    acq_act = AssessmentAcquisition(session=session)
    # test add acquisition object entity to graph associated with participant role NIDM_PARTICIPANT
    acq_entity = AssessmentObject(acquisition=acq_act)
    acq_entity.add_attributes(
        {Constants.NIDM["Q1"]: "Q1 Answer", Constants.NIDM["Q2"]: "Q2 Answer"}
    )
    # associate person as participant
    acq_act.add_qualified_association(person=person, role=Constants.NIDM_PARTICIPANT)

    # test add DemographicsAssessment acquisition activity / entity to graph and associate with session
    acq_act = AssessmentAcquisition(session=session)
    # test add acquisition object entity to graph associated with participant role NIDM_PARTICIPANT
    acq_entity = DemographicsObject(acquisition=acq_act)
    # add new person to graph
    person2 = acq_act.add_person(
        attributes={
            Constants.NIDM_FAMILY_NAME: "Doe",
            Constants.NIDM_GIVEN_NAME: "John",
        }
    )
    # associate person2 with assessment acquisition
    acq_act.add_qualified_association(person=person2, role=Constants.NIDM_PARTICIPANT)

    acq_entity.add_attributes({Constants.NIDM_AGE: 60, Constants.NIDM_GENDER: "Male"})

    # save a turtle file
    with open("test_nidm.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())

    # save a DOT graph as PDF
    project.save_DotGraph("test_nidm.png", format="png")


if __name__ == "__main__":
    main()



"""Tools for working with NIDM-Experiment files"""

import click
from rdflib import Graph, util
from nidm.core import Constants
from nidm.experiment.Query import GetParticipantIDs
from nidm.experiment.tools.click_base import cli


# adding click argument parsing
@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
@click.option(
    "--s",
    "-s",
    required=False,
    is_flag=True,
    help="If parameter set then files will be merged by ndar:src_subjec_id of prov:agents",
)
@click.option(
    "--out_file", "-o", required=True, help="File to write concatenated NIDM files"
)
def merge(nidm_file_list, s, out_file):
    """
    This function will merge NIDM files.  See command line parameters for supported merge operations.
    """

    # graph = Graph()
    # for nidm_file in nidm_file_list.split(','):
    #    graph.parse(nidm_file,format=util.guess_format(nidm_file))

    # create empty graph
    graph = Graph()
    # start with the first NIDM file and merge the rest into the first
    first = True
    for nidm_file in nidm_file_list.split(","):
        # if merging by subject:
        if s:
            if first:
                # get list of all subject IDs
                first_file_subjids = GetParticipantIDs([nidm_file])
                first = False
                first_graph = Graph()
                first_graph.parse(nidm_file, format=util.guess_format(nidm_file))
            else:
                # load second graph
                graph.parse(nidm_file, format=util.guess_format(nidm_file))

                # get list of second file subject IDs
                GetParticipantIDs([nidm_file])

                # for each UUID / subject ID look in graph and see if you can find the same ID.  If so get the UUID of
                # that prov:agent and change all the UUIDs in nidm_file to match then concatenate the two graphs.
                query = f"""

                    PREFIX prov:<http://www.w3.org/ns/prov#>
                    PREFIX sio: <http://semanticscience.org/ontology/sio.owl#>
                    PREFIX ndar: <https://ndar.nih.gov/api/datadictionary/v2/dataelement/>
                    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
                    PREFIX prov:<http://www.w3.org/ns/prov#>

                    SELECT DISTINCT ?uuid ?ID
                    WHERE {{

                        ?uuid a prov:Agent ;
                            {Constants.NIDM_SUBJECTID} ?ID .
                    FILTER(?ID =
                    """

                # add filters to above query to only look for subject IDs which are in the first file to merge into
                temp = True
                for ID in first_file_subjids["ID"]:
                    if temp:
                        query = query + '"' + ID + '"'
                        temp = False
                    else:
                        query = query + '|| ?ID= "' + ID + '"'

                query = query + ") }"

                qres = graph.query(query)

                # if len(qres) > 0 then we have matches so load the nidm_file into a temporary graph so we can
                # make changes to it then concatenate it.
                if len(qres) > 0:
                    # tmp = Graph()
                    # tmp.parse(nidm_file,format=util.guess_format(nidm_file))

                    # for each ID in the merged graph that matches an ID in the nidm_file graph
                    for row in qres:
                        # find ID from first file that matches ID in this file
                        t = first_file_subjids["ID"].str.match(row["ID"])
                        # then get uuid for that match from first file
                        uuid_replacement = first_file_subjids.iloc[
                            [*filter(t.get, t.index)][0], 0
                        ]

                        for s, p, o in graph.triples((None, None, None)):
                            if s == row["uuid"]:
                                # print(f"replacing subject in triple {s} {p} {o} with {uuid_to_replace}")
                                graph.add((uuid_replacement, p, o))
                                graph.remove((row["uuid"], p, o))
                            elif o == row["uuid"]:
                                # print(f"replacing object in triple {s} {p} {o} with {uuid_to_replace}")
                                graph.add((s, p, uuid_replacement))
                                graph.remove((s, p, row["uuid"]))
                            elif p == row["uuid"]:
                                # print(f"replacing predicate in triple {s} {p} {o} with {uuid_to_replace}")
                                graph.add((s, uuid_replacement, o))
                                graph.remove((s, row["uuid"], o))

                # merge updated graph

                graph = first_graph + graph

        graph.serialize(out_file, format="turtle")


if __name__ == "__main__":
    merge()



from argparse import ArgumentParser
from pprint import pprint
from nidm.experiment import Utils


def main():
    parser = ArgumentParser(
        description="This program will query SciCrunch term labels for query_string using key and print out the return JSON packet."
    )
    parser.add_argument(
        "-query_string", dest="query_string", required=True, help="Query String"
    )
    parser.add_argument(
        "-key", dest="key", required=True, help="SciCrunch API key to use for query"
    )
    args = parser.parse_args()

    # Test exact match search returning JSON package
    print("Testing term label search...")
    json_data = Utils.QuerySciCrunchTermLabel(args.key, args.query_string)
    print("Term label search returns:")
    print("-------------------------------------------")
    pprint(json_data)
    print("\n\n")

    # Test elastic search using CDEs and Terms + ancestors (simulates tagging sets of terms for NIDM use) returning JSON package
    print("Testing elastic search...")
    json_data = Utils.QuerySciCrunchElasticSearch(args.key, args.query_string)
    print("Elastic search returns:")
    print("-------------------------------------------")
    pprint(json_data)
    #  print("\n\n-------------------------------------------")
    #  print("Example terms listing from elastic search:")

    # example printing term label, definition, and preferred URL
    #  for term in json_data['hits']['hits']:
    #      #find preferred URL
    #      for items in term['_source']['existing_ids']:
    #          if items['preferred']=='1':
    #              preferred_url=items['iri']
    #      print(f"Label = {term['_source']['label']} \t Definition = {term['_source']['definition']} \t Preferred URL = {preferred_url} ")

    # example of uber elastic search query returns dictionary of label, definition, and preferred_url
    print("\n\n-------------------------------------------")
    print("Example uber elastic search:")
    results = Utils.GetNIDMTermsFromSciCrunch(args.key, args.query_string)
    for value in results.values():
        print(
            "Label: {label} \t Definition: {definition} \t Preferred URL: {preferred_url} ".format_map(
                value
            )
        )


if __name__ == "__main__":
    main()


# very simple test, just checking if main does not give any error
def test_main():
    main()



from nidm.experiment.Utils import fuzzy_match_terms_from_graph, load_nidm_owl_files


def test_loadowl():
    owl_graph = load_nidm_owl_files()
    fuzzy_match_terms_from_graph(owl_graph, "WisconsinCardSortingTest")



import prov.model as pm
from .Core import Core, getUUID
from ..core import Constants


class Project(pm.ProvActivity, Core):
    """Class for NIDM-Experiment Project-Level Objects.

    Default constructor uses empty graph with namespaces added from NIDM/Scripts/Constants.py.
    Additional alternate constructors for user-supplied graphs and default namespaces (i.e. from Constants.py)
    and user-supplied graph and namespaces

    @author: David Keator <dbkeator@uci.edu>
    @copyright: University of California, Irvine 2017

    """

    # constructor, adds project
    def __init__(
        self, attributes=None, empty_graph=False, uuid=None, add_default_type=True
    ):
        """
        Default constructor, creates document and adds Project activity to graph with optional attributes

        :param attributes: optional dictionary of attributes to add
        :empty_graph: if set to True, creates empty graph with no namespaces besides Prov defaults
        :uuid: if uuid is not None then use supplied uuid for project instead of generating one (for reading nidm docs)

        """

        if empty_graph:
            self.graph = Constants.NIDMDocument(namespaces=None)
        else:
            self.graph = Constants.NIDMDocument(namespaces=Constants.namespaces)

        if uuid is None:
            self._uuid = getUUID()

            # execute default parent class constructor
            super().__init__(
                self.graph,
                pm.QualifiedName(
                    pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                ),
                attributes,
            )
        else:
            self._uuid = uuid

            # check if niiri namespace is already defined
            niiri_ns = self.find_namespace_with_uri(uri=str(Constants.NIIRI))

            if niiri_ns is False:
                # execute default parent class constructor
                super().__init__(
                    self.graph,
                    pm.QualifiedName(
                        pm.Namespace("niiri", Constants.NIIRI), self.get_uuid()
                    ),
                    attributes,
                )
            else:
                # execute default parent class constructor
                super().__init__(
                    self.graph,
                    pm.QualifiedName(niiri_ns, self.get_uuid()),
                    attributes,
                )

        # add record to graph
        self.graph._add_record(self)
        # create empty sessions list
        self._sessions = []
        # create empty derivatives list
        self._derivatives = []
        # create empty data elements list
        self._dataelements = []

        if add_default_type:
            self.add_attributes({pm.PROV_TYPE: Constants.NIDM_PROJECT})

    def add_sessions(self, session):
        """
        Adds session to project, creating links and adding reference to sessions list

        :param session: object of type "Session" from nidm API
        :return true if session object added to project, false if session object is already in project

        """
        if session in self._sessions:
            return False
        else:
            # add session to self.sessions list
            self._sessions.extend([session])
            # get qname for niiri prefix and uuid...already in graph as we're adding a session and niiri added
            # when adding parent project
            niiri_qname = self.graph.valid_qualified_name("niiri:" + self.get_uuid())
            # create links in graph
            # session.add_attributes({str("dct:isPartOf"):self})
            if self.checkNamespacePrefix("dct"):
                dct_qname = self.graph.valid_qualified_name("dct:isPartOf")
                session.add_attributes({dct_qname: niiri_qname})

            else:
                session.add_attributes(
                    {
                        pm.QualifiedName(
                            pm.Namespace("dct", Constants.DCT), "isPartOf"
                        ): niiri_qname
                    }
                )
            return True

    def get_sessions(self):
        return self._sessions

    def get_derivatives(self):
        return self._derivatives

    def get_dataelements(self):
        return self._dataelements

    def add_derivatives(self, derivative):
        """
        Adds derivatives to project, creating links and adding reference to derivatives list
        :param derivative: object of type "Derivative" from nidm API
        :return true if derivative object added to project, false if derivative object is already in project
        """
        if derivative in self._derivatives:
            return False
        else:
            # add derivative to self._derivatives list
            self._derivatives.extend([derivative])
            # create links in graph
            # session.add_attributes({str("dct:isPartOf"):self})

            # when adding parent project
            niiri_qname = self.graph.valid_qualified_name("niiri:" + self.get_uuid())

            # check if DCT namespace is already added else add it
            if self.checkNamespacePrefix("dct"):
                dct_qname = self.graph.valid_qualified_name("dct:isPartOf")
                derivative.add_attributes({dct_qname: niiri_qname})
            else:
                derivative.add_attributes(
                    {
                        pm.QualifiedName(
                            pm.Namespace("dct", Constants.DCT), "isPartOf"
                        ): niiri_qname
                    }
                )
            return True

    def add_dataelements(self, dataelement):
        """
        Adds data elements to project, creating links and adding reference to data elements list
        :param dataelement: object of type "DataElement" from nidm API
        :return true if derivative object added to project, false if derivative object is already in project
        """
        if dataelement in self._dataelements:
            return False
        else:
            # add session to self.sessions list
            self._dataelements.extend([dataelement])
            # create links in graph
            # session.add_attributes({str("dct:isPartOf"):self})
            # dataelement.add_attributes({pm.QualifiedName(pm.Namespace("dct", Constants.DCT), 'isPartOf'): self})
            return True

    def __str__(self):
        return "NIDM-Experiment Project Class"

    sessions = property(get_sessions, add_sessions)
    derivatives = property(get_derivatives, add_derivatives)
    dataelements = property(get_dataelements, add_dataelements)



import hashlib
from os import environ, path
import pickle
import tempfile
from rdflib import Graph
from nidm.core import Constants
import nidm.experiment.Query
from nidm.util import urlretrieve


def download_cde_files():
    cde_dir = tempfile.gettempdir()

    for url in Constants.CDE_FILE_LOCATIONS:
        urlretrieve(url, f"{cde_dir}/{url.split('/')[-1]}")

    return cde_dir


def getCDEs(file_list=None):
    if getCDEs.cache:
        return getCDEs.cache

    hasher = hashlib.md5()
    hasher.update(str(file_list).encode("utf-8"))
    h = hasher.hexdigest()

    cache_file_name = tempfile.gettempdir() + f"/cde_graph.{h}.pickle"

    if path.isfile(cache_file_name):
        with open(cache_file_name, "rb") as fp:
            rdf_graph = pickle.load(fp)
        getCDEs.cache = rdf_graph
        return rdf_graph

    rdf_graph = Graph()

    if not file_list:
        cde_dir = ""
        if "CDE_DIR" in environ:
            cde_dir = environ["CDE_DIR"]

        if (not cde_dir) and (
            path.isfile("/opt/project/nidm/core/cde_dir/ants_cde.ttl")
        ):
            cde_dir = "/opt/project/nidm/core/cde_dir"

        if not cde_dir:
            cde_dir = download_cde_files()

        file_list = []
        for f in ["ants_cde.ttl", "fs_cde.ttl", "fsl_cde.ttl"]:
            fname = f"{cde_dir}/{f}"
            if path.isfile(fname):
                file_list.append(fname)

    for fname in file_list:
        if path.isfile(fname):
            cde_graph = nidm.experiment.Query.OpenGraph(fname)
            rdf_graph = rdf_graph + cde_graph

    with open(cache_file_name, "wb") as cache_file:
        pickle.dump(rdf_graph, cache_file)

    getCDEs.cache = rdf_graph
    return rdf_graph


getCDEs.cache = None



from argparse import ArgumentParser
from nidm.experiment.Utils import read_nidm


def main():
    parser = ArgumentParser()
    # parse command line arguments
    parser.add_argument(
        "-nidm", dest="nidm_file", required=True, help="NIDM-Exp RDF File to import"
    )
    parser.add_argument("-out", dest="outfile", required=True, help="output file name")
    args = parser.parse_args()

    project = read_nidm(args.nidm_file)

    print(f"Project: \n {project.get_uuid()}")
    sessions = project.get_sessions()
    print(f"Sessions:\n {sessions}")

    acquisitions = []
    for session in sessions:
        acquisitions = session.get_acquisitions()
        print(f"Acquisitions:\n {acquisitions}")

        for acq in acquisitions:
            acquisition_objects = acq.get_acquisition_objects()
            print(f"Acquisition Objects:\n {acquisition_objects}")

    # check for data elements
    print(f"Data Elements: \n {project.get_dataelements()}")

    # derivatives

    # and for derivatives
    print(f"Derivatives: \n {project.get_derivatives()}")
    for deriv in project.get_derivatives():
        derivobj = deriv.get_derivative_objects()
        print(f"Derivative Objects: \n {derivobj}")

    with open(args.outfile, "w", encoding="utf-8") as f:
        # serialize project for comparison with the original
        f.write(project.serializeTurtle())


if __name__ == "__main__":
    main()



from __future__ import annotations
from dataclasses import dataclass
import re
import pytest
from nidm.experiment import Navigate
from nidm.util import urlretrieve


@dataclass
class ProjectData:
    files: list[str]
    uri: str


@pytest.fixture(scope="module")
def brain_vol(brain_vol_files: list[str]) -> ProjectData:
    projects = Navigate.getProjects(brain_vol_files)
    project_uri = projects[0]
    return ProjectData(brain_vol_files, project_uri)


@pytest.fixture(scope="module")
def openneuro(tmp_path_factory: pytest.TempPathFactory) -> ProjectData:
    tmp_path = tmp_path_factory.mktemp("openneuro")
    urlretrieve(
        "https://raw.githubusercontent.com/ReproNim/simple2_NIDM_examples/refs/heads/master/datasets.datalad.org/openneuro/ds000110/nidm.ttl",
        tmp_path / "ds000110.nidm.ttl",
    )
    files = [str(tmp_path / "ds000110.nidm.ttl")]
    projects2 = Navigate.getProjects(files)
    project_uri = projects2[0]
    return ProjectData(files, project_uri)


def test_navigate_get_projects(brain_vol: ProjectData) -> None:
    projects = Navigate.getProjects(brain_vol.files)
    assert len(projects) == 2


def test_navigate_get_sessions(brain_vol: ProjectData) -> None:
    sessions = Navigate.getSessions(brain_vol.files, brain_vol.uri)
    assert len(sessions) > 0


def test_navigate_get_acquisitions_for_session(brain_vol: ProjectData) -> None:
    sessions = Navigate.getSessions(brain_vol.files, brain_vol.uri)
    for _ in sessions:
        acquisitions = Navigate.getAcquisitions(brain_vol.files, sessions[0])
        assert len(acquisitions) > 0


def test_navigate_get_subjects_for_acquisition(brain_vol: ProjectData) -> None:
    subjects = set()
    sessions = Navigate.getSessions(brain_vol.files, brain_vol.uri)
    for s in sessions:
        acquisitions = Navigate.getAcquisitions(brain_vol.files, s)
        for acq in acquisitions:
            sub = Navigate.getSubject(brain_vol.files, acq)
            assert sub is not None
            subjects.add(sub)
    assert len(subjects) > 5


def test_navigate_get_acquisition_data_by_session(openneuro: ProjectData) -> None:
    set_of_keys_returned = set()
    set_of_activities = set()

    sessions = Navigate.getSessions(openneuro.files, openneuro.uri)
    assert len(sessions) > 0
    for s in sessions:
        acquisitions = Navigate.getAcquisitions(openneuro.files, s)
        assert len(acquisitions) > 0
        for a in acquisitions:
            set_of_activities.add(str(a))
            ad = Navigate.getActivityData(openneuro.files, a)
            assert len(ad.data) > 5
            for vt in ad.data:
                set_of_keys_returned.add(vt.label)

    print(set_of_keys_returned)

    assert "age" in set_of_keys_returned
    assert "hadAcquisitionModality" in set_of_keys_returned


def test_navigate_get_acquisition_data_by_subject(openneuro: ProjectData) -> None:
    set_of_keys_returned = set()
    set_of_activities = set()

    subjects = Navigate.getSubjects(openneuro.files, openneuro.uri)
    assert len(subjects) > 0
    for s in subjects:
        activities = Navigate.getActivities(
            nidm_file_tuples=tuple(openneuro.files), subject_id=s
        )
        assert len(activities) > 0
        for a in activities:
            set_of_activities.add(str(a))
            ad = Navigate.getActivityData(openneuro.files, a)
            assert len(ad.data) > 5
            for vt in ad.data:
                set_of_keys_returned.add(vt.label)

    assert "age" in set_of_keys_returned
    assert "sex" in set_of_keys_returned
    assert "hadAcquisitionModality" in set_of_keys_returned
    assert "hadImageUsageType" in set_of_keys_returned


def test_navigate_get_sub_uuid_from_id(brain_vol: ProjectData) -> None:
    uuids = Navigate.getSubjectUUIDsfromID(
        nidm_file_tuples=brain_vol.files, sub_id="50653"
    )
    assert len(uuids) == 1
    assert re.match(
        "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", uuids[0]
    )  # check that it's a UUID



"""Tools for working with NIDM-Experiment files"""

import click
from nidm.experiment.Query import GetMergedGraph
from nidm.experiment.tools.click_base import cli


# adding click argument parsing
@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
@click.option(
    "--out_file", "-o", required=True, help="File to write concatenated NIDM files"
)
def concat(nidm_file_list, out_file):
    """
    This function will concatenate NIDM files.  Warning, no merging will be done so you may end up with
    multiple prov:agents with the same subject id if you're concatenating NIDM files from multiple visits of the
    same study.  If you want to merge NIDM files on subject ID see pynidm merge
    """
    # create empty graph
    graph = GetMergedGraph(nidm_file_list.split(","))
    graph.serialize(out_file, format="turtle")


if __name__ == "__main__":
    concat()
